"""Exec module for managing Amazon Neptune DB Cluster Parameter Group."""
from typing import Dict

__func_alias__ = {"list_": "list"}

RESOURCE_TYPE = "aws.neptune.db_cluster_parameter_group"


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict:
    """Retrieves the specified AWS Neptune DB Cluster Parameter Group.

    It also fetches the db cluster parameters for the supplied resource_id.
    Filters is not supported yet as per documentation for this resource on AWS.
    It returns idem object representation for db_cluster_parameter_group found if resource_id was specified.
    If resource_id wasn't specified, it returns the first resource found.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            DBClusterParameterGroupName of the resource in AWS.

    .. note::
        At the time of writing this module, filters parameter is not supported for this resource as per
        AWS documentation so idem-aws has omitted it.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The AWS Neptune DB Cluster Parameter Group in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.neptune.db_cluster_parameter_group.get name="idem_name" resource_id="resource_id"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.neptune.db_cluster_parameter_group.get(
                    ctx, name=name, resource_id=resource_id
                )
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.neptune.db_cluster_parameter_group.search_raw(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not ret["result"]:
        if "DBParameterGroupNotFoundFault" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type=RESOURCE_TYPE, name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["DBClusterParameterGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
        )
        return result

    resource = ret["ret"]["DBClusterParameterGroups"][0]
    resource_id = resource.get("DBClusterParameterGroupName")
    if len(ret["ret"]["DBClusterParameterGroups"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type=RESOURCE_TYPE, resource_id=resource_id
            )
        )
    arn = resource["DBClusterParameterGroupArn"]
    # get parameters
    parameters_raw = await hub.exec.boto3.client.neptune.describe_db_cluster_parameters(
        ctx=ctx, DBClusterParameterGroupName=resource_id
    )
    if not parameters_raw["result"]:
        result["result"] = False
        result["comment"].append(parameters_raw["comment"])
        return result
    parameters_raw = parameters_raw.get("ret")
    parameters = parameters_raw.get("Parameters")

    # get tags
    tags = await hub.tool.aws.neptune.tag.get_tags_for_resource(ctx, resource_arn=arn)
    if not tags["result"]:
        result["result"] = False
        result["comment"].append(tags["comment"])
        return result
    tags = tags["ret"]
    result[
        "ret"
    ] = await hub.tool.aws.neptune.conversion_utils.convert_raw_db_cluster_parameter_group_to_present(
        raw_resource=resource, idem_resource_name=name, tags=tags, parameters=parameters
    )
    return result


async def list_(hub, ctx, name) -> Dict:
    """Use a list of un-managed neptune db cluster parameter groups as a data-source.

    Args:
        name(str): The name of idem state.

    .. note::
        At the time of writing this module, filters parameter is not supported for this resource as per
        AWS documentation so idem-aws has omitted it.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The AWS Neptune DB Cluster Parameter Group in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.neptune.db_cluster_parameter_group.list name=name

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, **kwargs):
                ret = await hub.exec.aws.neptune.db_cluster_parameter_group.list(
                    ctx, name=name
                )
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.tool.aws.neptune.db_cluster_parameter_group.search_raw(
        ctx=ctx, name=name
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["DBClusterParameterGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
        )
        return result
    for db_cluster_parameter_group in ret["ret"]["DBClusterParameterGroups"]:
        db_cluster_parameter_group_name = db_cluster_parameter_group.get(
            "DBClusterParameterGroupName"
        )
        arn = db_cluster_parameter_group["DBClusterParameterGroupArn"]

        # get tags
        tags = await hub.tool.aws.neptune.tag.get_tags_for_resource(
            ctx, resource_arn=arn
        )
        if not tags["result"]:
            result["result"] = False
            result["comment"].append(tags["comment"])
            return result
        tags = tags["ret"]

        # get parameters
        parameters_raw = (
            await hub.exec.boto3.client.neptune.describe_db_cluster_parameters(
                ctx=ctx, DBClusterParameterGroupName=db_cluster_parameter_group_name
            )
        )
        if not parameters_raw["result"]:
            result["result"] = False
            result["comment"].append(parameters_raw["comment"])
            return result
        parameters_raw = parameters_raw.get("ret")
        parameters = parameters_raw.get("Parameters")

        result["ret"].append(
            await hub.tool.aws.neptune.conversion_utils.convert_raw_db_cluster_parameter_group_to_present(
                raw_resource=db_cluster_parameter_group,
                idem_resource_name=db_cluster_parameter_group_name,
                tags=tags,
                parameters=parameters,
            )
        )
    return result



CLI_CONFIG = {
    # pop-create options
    "services": {
        "subcommands": ["aws"],
        "dyne": "pop_create",
    },
}
CONFIG = {
    # pop-create options
    "services": {
        "default": [],
        "nargs": "*",
        "help": "The cloud services to target, defaults to all",
        "dyne": "pop_create",
    },
}
SUBCOMMANDS = {
    "aws": {
        "help": "Create idem_aws state modules by parsing boto3",
        "dyne": "pop_create",
    },
}
DYNE = {
    "acct": ["acct"],
    "exec": ["exec"],
    "pop_create": ["autogen"],
    "states": ["states"],
    "tool": ["tool"],
    "esm": ["esm"],
    "reconcile": ["reconcile"],
}



"""Exec module for managing Cloudwatch Log_Groups."""
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name,
    resource_id: str = None,
) -> Dict:
    """Get a log_group resource from AWS.

    If more than one resource is found, the first resource returned from AWS
    will be used. The function returns None when no resource is found.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str, optional):
            Aws logGroupNamePrefix (The prefix to match.)
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.cloudwatch.log_group.search_raw(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
    )
    if not ret["result"]:
        if "InvalidParameterException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.cloudwatch.log_group", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["logGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.cloudwatch.log_group", name=name
            )
        )
        return result

    resource = ret["ret"]["logGroups"][0]
    if len(ret["ret"]["logGroups"]) > 1:
        result["comment"].append(
            f"More than one aws.cloudwatch.log_group resource was found. Use resource {resource.get('logGroupName')}"
        )
    result[
        "ret"
    ] = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_log_group_to_present(
        ctx=ctx, raw_resource=resource, idem_resource_name=name
    )
    return result


async def list_(
    hub,
    ctx,
    name,
) -> Dict:
    """Fetch a list of log_group from AWS.

    The function returns empty list when no resource is found.

    Args:
        name(str):
            The name of the Idem state.
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.tool.aws.cloudwatch.log_group.search_raw(
        ctx=ctx,
        name=name,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["logGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.cloudwatch.log_group", name=name
            )
        )
        return result
    for log_group in ret["ret"]["logGroups"]:
        result["ret"].append(
            await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_log_group_to_present(
                ctx=ctx, raw_resource=log_group, idem_resource_name=name
            )
        )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS CloudFront to present input format.
"""


async def convert_raw_distribution_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    resource_id = raw_resource.get("Id")
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "arn": raw_resource.get("ARN"),
    }
    resource_parameters = OrderedDict(
        {
            "CallerReference": "caller_reference",
            "Comment": "comment",
            "Origins": "origins",
            "DefaultCacheBehavior": "default_cache_behaviour",
            "Logging": "logging",
            "Enabled": "enabled",
            "ViewerCertificate": "viewer_certificate",
            "Aliases": "aliases",
            "PriceClass": "price_class",
            "DefaultRootObject": "default_root_object",
            "OriginGroups": "origin_groups",
            "CacheBehaviors": "cache_behaviors",
            "CustomErrorResponses": "custom_error_responses",
            "Restrictions": "restrictions",
            "WebACLId": "web_acl_id",
            "HttpVersion": "http_version",
            "IsIPV6Enabled": "is_ipv6_enabled",
        }
    )

    distribution_config = raw_resource["DistributionConfig"]
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in distribution_config:
            resource_translated[parameter_present] = (
                dict(distribution_config.get(parameter_raw).copy())
                if isinstance(distribution_config.get(parameter_raw), dict)
                else distribution_config.get(parameter_raw)
            )

    if raw_resource.get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )
    if raw_resource.get("ETag") is not None:
        resource_translated["ETag"] = raw_resource.get("ETag")

    return resource_translated



from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, resource_id: str):
    """
    Get Document DB Subnet Group details from AWS given a AWS Document DB Subnet Group Name.

    Args:
        hub:
        ctx:
        resource_id(string): AWS Document DB Name

    """

    result = dict(comment=[], ret=None, result=True)
    before = await hub.tool.aws.docdb.db_subnet_group.search_raw(
        ctx, DBSubnetGroupName=resource_id
    )
    if not before["result"]:
        if "DBSubnetGroupNotFound" in str(before["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.docdb.db_subnet_group", name=resource_id
                )
            )
            result["comment"] += list(before["comment"])
            return result
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    if before["ret"]["DBSubnetGroups"]:
        resource = before["ret"]["DBSubnetGroups"]
        if len(resource) > 1:
            result["comment"].append(
                f"More than one aws.ec2.docdb.db_subnet_group resource was found. "
                f"Use resource {resource[0]['DBSubnetGroupName']}"
            )
        resource_arn = before["ret"]["DBSubnetGroups"][0].get("DBSubnetGroupArn")
        tags = await hub.tool.aws.docdb.tag.get_tags_for_resource(
            ctx, resource_arn=resource_arn
        )
        if not tags["result"]:
            result["result"] = False
            result["comment"] = tags["comment"]
            return result
        result[
            "ret"
        ] = hub.tool.aws.docdb.db_subnet_group.convert_raw_db_subnet_group_to_present(
            raw_resource=before["ret"]["DBSubnetGroups"][0],
            idem_resource_name=resource_id,
            tags=tags["ret"],
        )
    return result


async def list_(hub, ctx, DBSubnetGroupName: str = None, filters: List = None) -> Dict:
    """
    Use an un-managed DocDb Subnet as a data-source. Supply one of the inputs as the filter.

    Args:
        DBSubnetGroupName(string, optional): The name of the DBSubnetGroupName.
        default(bool, optional): Indicate whether the VPC is the default VPC.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/docdb.html#DocDB.Client.describe_db_subnet_groups

    """

    result = dict(comment=[], ret=[], result=True)
    ret = await hub.tool.aws.docdb.db_subnet_group.search_raw(
        ctx=ctx, filters=filters, DBSubnetGroupName=DBSubnetGroupName
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["DBSubnetGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.docdb.db_subnet_group", name=DBSubnetGroupName
            )
        )
        return result
    for subnetGroupName in ret["ret"]["DBSubnetGroups"]:
        name = subnetGroupName.get("DBSubnetGroupName")
        converted_db_group = (
            hub.tool.aws.docdb.db_subnet_group.convert_raw_db_subnet_group_to_present(
                raw_resource=subnetGroupName, idem_resource_name=name
            )
        )
        result["ret"].append(converted_db_group)
    return result



async def get(
    hub,
    ctx,
    resource_id: str,
    name: str = None,
):
    """
    Get an API Gateway Method Response from AWS.

    Args:
        hub:

        ctx:

        resource_id(string):
            Idem Resource ID that is generated once the resource is created,
                formatted as: rest_api_id-parent_resource_id-http_method-status_code.

        name(string, optional):
            An Idem name of the API Gateway Method.

    """

    result = dict(comment=[], ret=None, result=True)
    if resource_id and len(resource_id.split("-")) == 4:
        rest_api_id, parent_resource_id, http_method, status_code = resource_id.split(
            "-"
        )

        get_method_response = (
            await hub.exec.boto3.client.apigateway.get_method_response(
                ctx,
                restApiId=rest_api_id,
                resourceId=parent_resource_id,
                httpMethod=http_method,
                statusCode=status_code,
            )
        )

        if not get_method_response["result"]:
            if "NotFoundException" or "'NoneType' object is not iterable" in str(
                get_method_response["comment"]
            ):
                result["comment"].append(
                    hub.tool.aws.comment_utils.get_empty_comment(
                        resource_type="aws.apigateway.method_response", name=name
                    )
                )
                result["comment"] += list(get_method_response["comment"])
                return result
            result["comment"] += list(get_method_response["comment"])
            result["result"] = False
            return result

        if get_method_response["ret"]:
            result[
                "ret"
            ] = await hub.tool.aws.apigateway.method_response.convert_raw_method_response_to_present(
                idem_resource_name=name,
                raw_resource=get_method_response["ret"],
                resource_id=resource_id,
            )

        return result

    else:
        result["comment"] = (f"Invalid Resource ID '{resource_id}'.",)
        result["result"] = False
        return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.iam.create_instance_profile
hub.exec.boto3.client.iam.delete_instance_profile
hub.exec.boto3.client.iam.get_instance_profile
hub.exec.boto3.client.iam.list_instance_profiles
hub.exec.boto3.client.iam.tag_instance_profile
hub.exec.boto3.client.iam.untag_instance_profile
resource = hub.tool.boto3.resource.create(ctx, "iam", "InstanceProfile", name)
hub.tool.boto3.resource.exec(resource, add_role, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, remove_role, *args, **kwargs)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


TREQ = {
    "absent": {
        "require": [
            "aws.iam.role.absent",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    path: str = None,
    roles: List = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

     Creates a new instance profile. For information about instance profiles, see Using roles for applications on
    Amazon EC2 in the IAM User Guide, and Instance profiles in the Amazon EC2 User Guide.  For information about the
    number of instance profiles you can create, see IAM object quotas in the IAM User Guide.
    Tags can be set on a newly created instance profile, or updated on an existing one.
    Roles are set and updated as a day-2 operation.

    Args:
        name(Text): AWS Instance Profile Name.
        resource_id(Text, Optional): AWS Instance Profile Name.
        path(Text, Optional): The path to the instance profile.
        roles(Text, Optional): A list with up to one role name. Instance profile can be associated with a single role.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the IAM instance profile. Each tag consists of
            a key name and an associated value. Defaults to None.
            * (Key): The key name that can be used to look up or retrieve the associated value. For example,
                Department or Cost Center are common choices.
            * (Value): The value associated with this tag. For example, tags with a key name of Department could have
                values such as Human Resources, Accounting, and Support. Tags with a key name of Cost Center
                might have values that consist of the number associated with the different cost centers in your
                company. Typically, many resources have tags with the same key name but with different values.
                Amazon Web Services always interprets the tag Value as a single string. If you need to store an
                array, you can store comma-separated values in the string. However, you must interpret the value
                in your code.

    Returns:
        Dict[str, Any]

    Request Syntax:
      [iam-instance-profile-name]:
        aws.iam.instance_profile.present:
          - name: 'string'
          - resource_id: 'string'
          - path: 'string'
          - roles:
            - RoleName: 'string'
          - Tags:
            - Key: 'string'
              Value: 'string'

    Example:

        .. code-block:: sls

        inst-profile-1:
          aws.iam.instance_profile.present:
          - name: inst-profile-1
          - path: /aws-service-inst-profile/
          - roles:
            - RoleName: role-1
          - Tags:
            - Key: test-1
              Value: test-1-val

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.aws.iam.instance_profile.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if before["ret"]:
        result["old_state"] = before["ret"]
        result["new_state"] = copy.deepcopy(before["ret"])
        result["comment"] = (f"'{name}' already exists",)
        # Update key tags if tags are specified
        if tags is not None and tags != result["old_state"].get("tags", None):
            hub.log.debug(f"aws.iam.instance_profile '{name}' tags update")
            update_ret = (
                await hub.exec.aws.iam.instance_profile.update_instance_profile_tags(
                    ctx=ctx,
                    instance_profile_name=resource_id if resource_id else name,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
            )

            result["result"] = update_ret["result"]
            result["comment"] = result["comment"] + update_ret["comment"]
            if not result["result"]:
                return result
            if ctx.get("test", False) and update_ret["ret"] is not None:
                result["new_state"]["tags"] = update_ret["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "path": path,
                    "roles": roles,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.iam.instance_profile", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.iam.create_instance_profile(
                ctx,
                InstanceProfileName=name,
                Path=path,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not ret["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.iam.instance_profile", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    update_ret = None
    if roles is not None:
        # Instance profile can be associated only with a single role
        # Remove existing role
        curr_role = None
        new_role = None
        if before["ret"] and before["ret"].get("roles", []):
            curr_role = before["ret"].get("roles")[0]["RoleName"]
        if roles:
            new_role = roles[0]["RoleName"]
        if curr_role != new_role:
            if curr_role is not None and not ctx.get("test", False):
                hub.log.debug(
                    f"Removing role '{curr_role}' from instance profile '{name}'"
                )
                update_ret = (
                    await hub.exec.boto3.client.iam.remove_role_from_instance_profile(
                        ctx, InstanceProfileName=name, RoleName=curr_role
                    )
                )
                if not update_ret["result"]:
                    result["result"] = False
                    result["comment"] = update_ret["comment"]
                    return result

                result["comment"] = result["comment"] + (
                    f"Role '{curr_role}' was removed from instance profile '{name}'",
                )

            if new_role is not None and not ctx.get("test", False):
                hub.log.debug(f"Adding role '{new_role}' to instance profile '{name}'")
                update_ret = (
                    await hub.exec.boto3.client.iam.add_role_to_instance_profile(
                        ctx, InstanceProfileName=name, RoleName=new_role
                    )
                )
                if not update_ret["result"]:
                    result["result"] = False
                    result["comment"] = update_ret["comment"]
                    return result

                result["comment"] = result["comment"] + (
                    f"Role '{new_role}' was added to instance profile '{name}'",
                )
            # Test context when updating roles
            if ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Updated role in instance profile '{name}'.",
                )
                result["new_state"]["roles"] = (
                    [{"RoleName": new_role}] if new_role is not None else []
                )
                return result

    if ctx.get("test", False):
        return result

    if before["ret"] and update_ret is None:
        result["new_state"] = copy.deepcopy(result["old_state"])
    else:
        after = await hub.exec.aws.iam.instance_profile.get(
            ctx, name=name, resource_id=resource_id
        )
        if not after["result"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result
        result["new_state"] = after["ret"]

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified instance profile. The instance profile must not have an associated role.  Make sure that
    you do not have any Amazon EC2 instances running with the instance profile you are about to delete. Deleting a
    role or instance profile that is associated with a running instance will break any applications running on the
    instance.  For more information about instance profiles, see About instance profiles.

    Args:
        name(Text): AWS Instance Profile Name.
        resource_id(Text, Optional): AWS Instance Profile Name. If not specified, Idem will use "name"
         parameter to identify the role policy on AWS.

    Returns:
        Dict[str, Any]

    Request Syntax:
      [iam-instance-profile-name]:
        aws.iam.instance_profile.present:
          - name: 'string'

    Examples:

        .. code-block:: sls

            instance-profile-test:
              aws.iam.instance_profile.absent:
                - name: instance-profile-test
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.aws.iam.instance_profile.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.instance_profile", name=name
        )
        return result
    else:
        try:
            result["old_state"] = before["ret"]
            if ctx.get("test", False):
                result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                    resource_type="aws.iam.instance_profile", name=name
                )
                return result

            ret = await hub.exec.boto3.client.iam.delete_instance_profile(
                ctx,
                InstanceProfileName=name,
            )
            if not ret["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.iam.instance_profile", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the instance profiles that have the specified path prefix. If there are none, the operation returns an
    empty list. For more information about instance profiles, see About instance profiles.  IAM resource-listing
    operations return a subset of the available attributes for the resource. For example, this operation does not
    return tags, even though they are an attribute of the returned object. To view all of the information for an
    instance profile, see GetInstanceProfile.  You can paginate the results using the MaxItems and Marker
    parameters.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.iam.instance_profile
    """

    result = {}
    ret = await hub.exec.boto3.client.iam.list_instance_profiles(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.iam.instance_profile {ret['comment']}")
        return {}

    for instance_profile in ret["ret"]["InstanceProfiles"]:
        # This is required to get tags for each instance profile list_instance_profile_roles is not working
        resource = await hub.exec.aws.iam.instance_profile.get(
            ctx,
            name=instance_profile["InstanceProfileName"],
            resource_id=instance_profile["InstanceProfileName"],
        )
        if not resource["result"]:
            hub.log.info(
                f"Failed to retrieve aws.iam.instance_profile {instance_profile['InstanceProfileName']}: {resource['comment']}"
            )
            continue

        resource_key = f"iam-instance-profile-{resource['ret']['name']}"
        result[resource_key] = {
            "aws.iam.instance_profile.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource["ret"].items()
            ]
        }

    return result


async def search(hub, ctx, name, resource_id: str = None):
    """
    Provides details about a specific instance profile as a data-source. This function has been deprecated.
    Please use exec.run with aws.iam.instance_profile.get instead.

    Args:
        name(Text): AWS Instance Profile name to identify the resource.
        resource_id(Text, Optional): AWS Instance Profile name to identify the resource.

    Request Syntax:
        [Idem-state-name]:
           aws.iam.instance_profile.search:
              - name: 'string'
              - resource_id: 'string'

    Response Syntax:
        The following fields will be returned to the `new_state` of the response.

          - name: 'string'
          - resource_id: 'string'
          - arn: 'string'
          - path: 'string'
          - roles:
            - RoleName: 'string'
            - RoleArn: 'string'
            - RoleId: 'string'
          - Tags:
            - Key: 'string'
              Value: 'string'

    Examples:

        Input state file:

        .. code-block:: sls

            instance-profile-test-search:
              aws.iam.instance_profile.search:
                - name: instance-profile-test-search
                - resource_id: instance-profile-test-search

        Sample response:
              - name: inst-profile-1
              - resource_id: inst-profile-1
              - arn:
              - path: /aws-service-inst-profile/
              - roles:
                - RoleName: role-1
                - RoleArn: arn:aws:iam::000000000000:role/idem-test-role-09e61fc9-d078-48a6-b297-8969a509c7cd
                - RoleId: tpx1mo6cjpsvh95ex5mq
              - Tags:
                - Key: test-1
                  Value: test-1-val
    """

    hub.log.warning(
        f"aws.iam.instance_profile.search '{name}' state has been deprecated. Please use exec.run with aws.iam.instance_profile.get instead."
    )
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    ret = await hub.exec.aws.iam.instance_profile.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    result["result"] = ret["result"]
    result["comment"] = ret["comment"]
    if result["result"]:
        result["old_state"] = ret["ret"]
        # Populate both "old_state" and "new_state" with the same data
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
 Utility functions to support aws efs file-system mount target operations
"""


def convert_raw_mount_target_to_present(
    hub, ctx, mount_target: Dict[str, Any], security_groups: []
) -> Dict[str, Any]:
    result = {}
    if mount_target is None:
        return None

    resource_parameters = OrderedDict(
        {
            "FileSystemId": "file_system_id",
            "SubnetId": "subnet_id",
            "IpAddress": "ip_address",
        }
    )
    # Populate state parameters
    result["name"] = mount_target.get("MountTargetId")
    result["resource_id"] = mount_target.get("MountTargetId")
    result["security_groups"] = security_groups

    for parameter_key, parameter_value in resource_parameters.items():
        if parameter_key in mount_target:
            result[parameter_value] = mount_target.get(parameter_key)

    return result



def get(hub, ctx) -> str:
    """
    Get the region name from ctx and fall back to the region in config
    """
    if ctx.acct.get("region_name"):
        hub.log.trace("Using region name from acct profile")
        return ctx.acct["region_name"]

    opt = getattr(hub, "OPT") or {}
    acct = opt.get("acct") or {}
    extras = acct.get("extras") or {}
    aws_opts = extras.get("aws") or {}
    if aws_opts.get("region_name"):
        hub.log.trace("Using region name from idem config")
        return aws_opts["region_name"]

    hub.log.trace(
        "No acct profile specified, region and auth tokens may be used by boto from awscli config"
    )



from typing import Dict
from typing import List

from dict_tools import data


def get_notification_configuration_diff(
    hub, new_notifications: Dict[str, List], old_state: Dict[str, List]
) -> Dict[str, List]:
    """
    This functions helps in comparing two dicts.
    It compares each key value in both the dicts and return diff notifications configuration

    Returns:
        {Resultant notifications dictionary}

    """

    old_state_notification = old_state.get("notifications")

    old_state_notification = (
        old_state_notification if old_state_notification is not None else {}
    )

    return data.recursive_diff(
        new_notifications, old_state_notification, ignore_order=True
    )


def get_create_bucket_payload(
    hub,
    acl,
    name,
    create_bucket_configuration,
    grant_full_control,
    grant_read,
    grant_read_acp,
    grant_write,
    grant_write_acp,
    object_lock_enabled_for_bucket,
    object_ownership,
) -> Dict[str, object]:
    """
    This functions helps in creating payload to create bucket.

    Returns:
        {payload dictionary}

    """
    payload = {"Bucket": name}
    if acl is not None:
        payload["ACL"] = acl
    if create_bucket_configuration is not None:
        payload["CreateBucketConfiguration"] = create_bucket_configuration
    if grant_full_control is not None:
        payload["GrantFullControl"] = grant_full_control
    if grant_read is not None:
        payload["GrantRead"] = grant_read
    if grant_read_acp is not None:
        payload["GrantReadACP"] = grant_read_acp
    if grant_write is not None:
        payload["GrantWrite"] = grant_write
    if grant_write_acp is not None:
        payload["GrantWriteACP"] = grant_write_acp
    if object_lock_enabled_for_bucket is not None:
        payload["ObjectLockEnabledForBucket"] = object_lock_enabled_for_bucket
    if object_ownership is not None:
        payload["ObjectOwnership"] = object_ownership
    return payload


def get_grantee_payload(hub, permission, grantee) -> str:
    """
    This functions helps in creating payload of grantee.

    Returns:
        {grantee string}

    """
    if permission:
        delimiter = ", "
    else:
        delimiter = ""
    if "ID" in grantee.keys():
        permission += delimiter + "ID=" + grantee["ID"]
    if "EmailAddress" in grantee.keys():
        permission += delimiter + "EmailAddress=" + grantee["EmailAddress"]
    if "URI" in grantee.keys():
        permission += delimiter + "URI=" + grantee["URI"]
    return permission



"""Exec module for managing Scaling policies."""
from typing import Dict


async def get(
    hub,
    ctx,
    name,
    auto_scaling_group_name: str,
    resource_id: str = None,
    policy_type: str = None,
) -> Dict:
    """Retrieves the specified scaling policy.

    Use an un-managed scaling policy as a data-source. Supply one of the inputs as the filter.
    If resource_id is set, autoscaling group is searched based on the resource_id and policy_type is ignored.

    Args:
        name(str):
            The name of the Idem state.

        auto_scaling_group_name(str):
            The name of the Auto Scaling group.

        resource_id(str, Optional):
            Name of the Auto Scaling group's policy.

        policy_type(str, Optional):
            Type of Auto Scaling group's policy. This is used to fetch a policy if resource_id is not specified.

    Returns:
        .. code-block:: python

            {"result": True|False, "comment": list, "ret": None|dict}

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.autoscaling.scaling_policy.get auto_scaling_group_name="asg_1" resource_id="resource_id"

        Calling this exec module function from within a state module in pure python

        .. code-block:: python

            async def state_function(hub, ctx, name, auto_scaling_group_name, resource_id):
                ret = await hub.exec.aws.autoscaling.scaling_policy.get(
                    ctx=ctx,
                    name=name,
                    resource_id=resource_id,
                    auto_scaling_group_name=auto_scaling_group_name
                )

    """
    result = dict(comment=[], ret=None, result=True)
    if resource_id:
        ret = await hub.exec.boto3.client.autoscaling.describe_policies(
            ctx=ctx,
            AutoScalingGroupName=auto_scaling_group_name,
            PolicyNames=[resource_id],
        )
    else:
        ret = await hub.exec.boto3.client.autoscaling.describe_policies(
            ctx=ctx,
            AutoScalingGroupName=auto_scaling_group_name,
            PolicyTypes=[policy_type] if policy_type else None,
        )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["ScalingPolicies"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.autoscaling.scaling_policy", name=name
            )
        )
        return result

    resource = ret["ret"]["ScalingPolicies"][0]
    if len(ret["ret"]["ScalingPolicies"]) > 1:
        result["comment"].append(
            f"More than one aws.autoscaling.scaling_policy resource was found. Use resource {resource.get('PolicyName')}"
        )
    result[
        "ret"
    ] = hub.tool.aws.autoscaling.conversion_utils.convert_raw_scaling_policy_to_present(
        ctx, raw_resource=resource
    )
    return result



from typing import Any
from typing import Dict


async def update_scalable_target(
    hub,
    ctx,
    name: str,
    before: Dict[str, Any],
    service_namespace: str,
    scaling_resource_id: str,
    scalable_dimension: str,
    min_capacity: int,
    max_capacity: int,
    role_arn: str,
    suspended_state: Dict[str, bool],
):
    result = dict(comment=(), result=True, ret=None)
    update_allowed_params = [
        "min_capacity",
        "max_capacity",
        "role_arn",
        "suspended_state",
    ]

    modified_params = {}
    for parameter in update_allowed_params:
        if (
            locals()[parameter] is not None
            and before.get(parameter) != locals()[parameter]
        ):
            modified_params[parameter] = locals()[parameter]
    if modified_params:
        result["ret"] = modified_params
        if not ctx.get("test", False):
            ret = await hub.exec.boto3.client[
                "application-autoscaling"
            ].register_scalable_target(
                ctx,
                ServiceNamespace=service_namespace,
                ResourceId=scaling_resource_id,
                ScalableDimension=scalable_dimension,
                MinCapacity=min_capacity,
                MaxCapacity=max_capacity,
                RoleARN=role_arn,
                SuspendedState=suspended_state,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.application_autoscaling.scalable_target", name=name
            )
    return result



from typing import Dict


async def search_raw(hub, ctx, name, resource_id: str = None) -> Dict:
    """
    Fetch one or more log groups from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str, optional):
            Aws logGroupNamePrefix (The prefix to match.)

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.logs.describe_log_groups(
        ctx,
        logGroupNamePrefix=resource_id if resource_id else None,
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result



"""State module for managing Amazon Cloudfront Distribution."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    caller_reference: str,
    origins: make_dataclass(
        "Origins",
        [
            ("Quantity", int),
            (
                "Items",
                List[
                    make_dataclass(
                        "Origin",
                        [
                            ("Id", str),
                            ("DomainName", str),
                            ("OriginPath", str, field(default=None)),
                            (
                                "CustomHeaders",
                                make_dataclass(
                                    "CustomHeaders",
                                    [
                                        ("Quantity", int),
                                        (
                                            "Items",
                                            List[
                                                make_dataclass(
                                                    "OriginCustomHeader",
                                                    [
                                                        ("HeaderName", str),
                                                        ("HeaderValue", str),
                                                    ],
                                                )
                                            ],
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "S3OriginConfig",
                                make_dataclass(
                                    "S3OriginConfig", [("OriginAccessIdentity", str)]
                                ),
                                field(default=None),
                            ),
                            (
                                "CustomOriginConfig",
                                make_dataclass(
                                    "CustomOriginConfig",
                                    [
                                        ("HTTPPort", int),
                                        ("HTTPSPort", int),
                                        ("OriginProtocolPolicy", str),
                                        (
                                            "OriginSslProtocols",
                                            make_dataclass(
                                                "OriginSslProtocols",
                                                [
                                                    ("Quantity", int),
                                                    ("Items", List[str]),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                        ("OriginReadTimeout", int, field(default=None)),
                                        (
                                            "OriginKeepaliveTimeout",
                                            int,
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("ConnectionAttempts", int, field(default=None)),
                            ("ConnectionTimeout", int, field(default=None)),
                            (
                                "OriginShield",
                                make_dataclass(
                                    "OriginShield",
                                    [
                                        ("Enabled", bool),
                                        (
                                            "OriginShieldRegion",
                                            str,
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                        ],
                    )
                ],
            ),
        ],
    ),
    default_cache_behaviour: make_dataclass(
        "DefaultCacheBehavior",
        [
            ("TargetOriginId", str),
            ("ViewerProtocolPolicy", str),
            (
                "TrustedSigners",
                make_dataclass(
                    "TrustedSigners",
                    [
                        ("Enabled", bool),
                        ("Quantity", int),
                        ("Items", List[str], field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "TrustedKeyGroups",
                make_dataclass(
                    "TrustedKeyGroups",
                    [
                        ("Enabled", bool),
                        ("Quantity", int),
                        ("Items", List[str], field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "AllowedMethods",
                make_dataclass(
                    "AllowedMethods",
                    [
                        ("Quantity", int),
                        ("Items", List[str]),
                        (
                            "CachedMethods",
                            make_dataclass(
                                "CachedMethods",
                                [("Quantity", int), ("Items", List[str])],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            ("SmoothStreaming", bool, field(default=None)),
            ("Compress", bool, field(default=None)),
            (
                "LambdaFunctionAssociations",
                make_dataclass(
                    "LambdaFunctionAssociations",
                    [
                        ("Quantity", int),
                        (
                            "Items",
                            List[
                                make_dataclass(
                                    "LambdaFunctionAssociation",
                                    [
                                        ("LambdaFunctionARN", str),
                                        ("EventType", str),
                                        ("IncludeBody", bool, field(default=None)),
                                    ],
                                )
                            ],
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            (
                "FunctionAssociations",
                make_dataclass(
                    "FunctionAssociations",
                    [
                        ("Quantity", int),
                        (
                            "Items",
                            List[
                                make_dataclass(
                                    "FunctionAssociation",
                                    [("FunctionARN", str), ("EventType", str)],
                                )
                            ],
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            ("FieldLevelEncryptionId", str, field(default=None)),
            ("RealtimeLogConfigArn", str, field(default=None)),
            ("CachePolicyId", str, field(default=None)),
            ("OriginRequestPolicyId", str, field(default=None)),
            ("ResponseHeadersPolicyId", str, field(default=None)),
            (
                "ForwardedValues",
                make_dataclass(
                    "ForwardedValues",
                    [
                        ("QueryString", bool),
                        (
                            "Cookies",
                            make_dataclass(
                                "CookiePreference",
                                [
                                    ("Forward", str),
                                    (
                                        "WhitelistedNames",
                                        make_dataclass(
                                            "CookieNames",
                                            [
                                                ("Quantity", int),
                                                (
                                                    "Items",
                                                    List[str],
                                                    field(default=None),
                                                ),
                                            ],
                                        ),
                                        field(default=None),
                                    ),
                                ],
                            ),
                        ),
                        (
                            "Headers",
                            make_dataclass(
                                "Headers",
                                [
                                    ("Quantity", int),
                                    ("Items", List[str], field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                        (
                            "QueryStringCacheKeys",
                            make_dataclass(
                                "QueryStringCacheKeys",
                                [
                                    ("Quantity", int),
                                    ("Items", List[str], field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            ("MinTTL", int, field(default=None)),
            ("DefaultTTL", int, field(default=None)),
            ("MaxTTL", int, field(default=None)),
        ],
    ),
    comment: str,
    enabled: bool,
    logging: make_dataclass(
        "LoggingConfig",
        [("Enabled", bool), ("IncludeCookies", bool), ("Bucket", str), ("Prefix", str)],
    ) = None,
    viewer_certificate: make_dataclass(
        "ViewerCertificate",
        [
            ("CloudFrontDefaultCertificate", bool, field(default=None)),
            ("IAMCertificateId", str, field(default=None)),
            ("ACMCertificateArn", str, field(default=None)),
            ("SSLSupportMethod", str, field(default=None)),
            ("MinimumProtocolVersion", str, field(default=None)),
            ("Certificate", str, field(default=None)),
            ("CertificateSource", str, field(default=None)),
        ],
    ) = None,
    aliases: make_dataclass(
        "Aliases", [("Quantity", int), ("Items", List[str], field(default=None))]
    ) = None,
    price_class: str = None,
    default_root_object: str = None,
    origin_groups: make_dataclass(
        "OriginGroups",
        [
            ("Quantity", int),
            (
                "Items",
                List[
                    make_dataclass(
                        "OriginGroup",
                        [
                            ("Id", str),
                            (
                                "FailoverCriteria",
                                make_dataclass(
                                    "OriginGroupFailoverCriteria",
                                    [
                                        (
                                            "StatusCodes",
                                            make_dataclass(
                                                "StatusCodes",
                                                [
                                                    ("Quantity", int),
                                                    ("Items", List[int]),
                                                ],
                                            ),
                                        )
                                    ],
                                ),
                            ),
                            (
                                "Members",
                                make_dataclass(
                                    "OriginGroupMembers",
                                    [
                                        ("Quantity", int),
                                        (
                                            "Items",
                                            List[
                                                make_dataclass(
                                                    "OriginGroupMember",
                                                    [("OriginId", str)],
                                                )
                                            ],
                                        ),
                                    ],
                                ),
                            ),
                        ],
                    )
                ],
                field(default=None),
            ),
        ],
    ) = None,
    cache_behaviors: make_dataclass(
        "CacheBehaviors",
        [
            ("Quantity", int),
            (
                "Items",
                List[
                    make_dataclass(
                        "CacheBehavior",
                        [
                            ("PathPattern", str),
                            ("TargetOriginId", str),
                            ("ViewerProtocolPolicy", str),
                            (
                                "TrustedSigners",
                                make_dataclass(
                                    "TrustedSigners",
                                    [
                                        ("Enabled", bool),
                                        ("Quantity", int),
                                        ("Items", List[str], field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "TrustedKeyGroups",
                                make_dataclass(
                                    "TrustedKeyGroups",
                                    [
                                        ("Enabled", bool),
                                        ("Quantity", int),
                                        ("Items", List[str], field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "AllowedMethods",
                                make_dataclass(
                                    "AllowedMethods",
                                    [
                                        ("Quantity", int),
                                        ("Items", List[str]),
                                        (
                                            "CachedMethods",
                                            make_dataclass(
                                                "CachedMethods",
                                                [
                                                    ("Quantity", int),
                                                    ("Items", List[str]),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("SmoothStreaming", bool, field(default=None)),
                            ("Compress", bool, field(default=None)),
                            (
                                "LambdaFunctionAssociations",
                                make_dataclass(
                                    "LambdaFunctionAssociations",
                                    [
                                        ("Quantity", int),
                                        (
                                            "Items",
                                            List[
                                                make_dataclass(
                                                    "LambdaFunctionAssociation",
                                                    [
                                                        ("LambdaFunctionARN", str),
                                                        ("EventType", str),
                                                        (
                                                            "IncludeBody",
                                                            bool,
                                                            field(default=None),
                                                        ),
                                                    ],
                                                )
                                            ],
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "FunctionAssociations",
                                make_dataclass(
                                    "FunctionAssociations",
                                    [
                                        ("Quantity", int),
                                        (
                                            "Items",
                                            List[
                                                make_dataclass(
                                                    "FunctionAssociation",
                                                    [
                                                        ("FunctionARN", str),
                                                        ("EventType", str),
                                                    ],
                                                )
                                            ],
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("FieldLevelEncryptionId", str, field(default=None)),
                            ("RealtimeLogConfigArn", str, field(default=None)),
                            ("CachePolicyId", str, field(default=None)),
                            ("OriginRequestPolicyId", str, field(default=None)),
                            ("ResponseHeadersPolicyId", str, field(default=None)),
                            (
                                "ForwardedValues",
                                make_dataclass(
                                    "ForwardedValues",
                                    [
                                        ("QueryString", bool),
                                        (
                                            "Cookies",
                                            make_dataclass(
                                                "CookiePreference",
                                                [
                                                    ("Forward", str),
                                                    (
                                                        "WhitelistedNames",
                                                        make_dataclass(
                                                            "CookieNames",
                                                            [
                                                                ("Quantity", int),
                                                                (
                                                                    "Items",
                                                                    List[str],
                                                                    field(default=None),
                                                                ),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                ],
                                            ),
                                        ),
                                        (
                                            "Headers",
                                            make_dataclass(
                                                "Headers",
                                                [
                                                    ("Quantity", int),
                                                    (
                                                        "Items",
                                                        List[str],
                                                        field(default=None),
                                                    ),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                        (
                                            "QueryStringCacheKeys",
                                            make_dataclass(
                                                "QueryStringCacheKeys",
                                                [
                                                    ("Quantity", int),
                                                    (
                                                        "Items",
                                                        List[str],
                                                        field(default=None),
                                                    ),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("MinTTL", int, field(default=None)),
                            ("DefaultTTL", int, field(default=None)),
                            ("MaxTTL", int, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
        ],
    ) = None,
    custom_error_responses: make_dataclass(
        "CustomErrorResponses",
        [
            ("Quantity", int),
            (
                "Items",
                List[
                    make_dataclass(
                        "CustomErrorResponse",
                        [
                            ("ErrorCode", int),
                            ("ResponsePagePath", str, field(default=None)),
                            ("ResponseCode", str, field(default=None)),
                            ("ErrorCachingMinTTL", int, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
        ],
    ) = None,
    restrictions: make_dataclass(
        "Restrictions",
        [
            (
                "GeoRestriction",
                make_dataclass(
                    "GeoRestriction",
                    [
                        ("RestrictionType", str),
                        ("Quantity", int),
                        ("Items", List[str], field(default=None)),
                    ],
                ),
            )
        ],
    ) = None,
    web_acl_id: str = None,
    http_version: str = None,
    is_ipv6_enabled: bool = True,
    resource_id: str = None,
    tags: Dict[str, str] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=60)),
                        ("max_attempts", int, field(default=35)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=60)),
                        ("max_attempts", int, field(default=35)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""Creates a new web distribution.

    You create a CloudFront distribution to tell CloudFront where you want content
    to be delivered from, and the details about how to track and manage content delivery. Send a POST request to the
    /CloudFront API version/distribution/distribution ID resource.

    .. Warning::
        When you update a distribution, there are more
        required fields than when you create a distribution. When you update your distribution by using
        UpdateDistribution, follow the steps included in the documentation to get the current configuration and then
        make your updates. This helps to make sure that you include all the required fields. To view a summary, see
        Required Fields for Create Distribution and Update Distribution in the Amazon CloudFront Developer Guide.

    Args:
        name(str):
            An Idem name of the resource.

        caller_reference (str):
            A unique value (for example, a date-time stamp) that ensures that the request can't be replayed.

            If the value of CallerReference is new (regardless of the content of the DistributionConfig
            object), CloudFront creates a new distribution.

            If CallerReference is a value that you already sent in a previous request to create a distribution,
            CloudFront returns a DistributionAlreadyExists error.

        resource_id(str, Optional):
            AWS cloudfront distribution ID.

        aliases(dict, Optional):
            A complex type that contains information about CNAME (alternate domain names),
            if any, for this distribution.

            * Quantity (int):
              The number of CNAME aliases, if any, that you want to associate with this distribution.
            * Items (list[str], Optional):
              A complex type that contains the CNAME aliases, if any, that you want to associate with this

        default_root_object (str, Optional):
            The object that you want CloudFront to request from your origin (for example, index.html) when a
            viewer requests the root URL for your distribution (http://www.example.com) instead of an object
            in your distribution (http://www.example.com/product-description.html).
            Specifying a default root object avoids exposing the contents of your distribution.

            Specify only the object name, for example, index.html. Don't add a / before the object name.

            If you don't want to specify a default root object when you create a distribution,
            include an empty DefaultRootObject element.

            To delete the default root object from an existing distribution, update the distribution
            configuration and include an empty DefaultRootObject element.

            To replace the default root object, update the distribution configuration and specify the new object.

            For more information about the default root object, see Creating a Default Root Object in the
            Amazon CloudFront Developer Guide.

        origins(dict):
            A complex type that contains information about origins for this distribution.
                * Quantity (int): The number of origins for this distribution.
                * Items (list[dict[str, Any]]): A list of origins.
                    * Id (str): A unique identifier for the origin. This value must be unique within the distribution.
                      Use this value to specify the TargetOriginId in a CacheBehavior or DefaultCacheBehavior.
                    * DomainName (str): The domain name for the origin.
                      For more information, see Origin Domain Name in the Amazon CloudFront Developer Guide.
                    * OriginPath (str, Optional):
                      An optional path that CloudFront appends to the origin domain name when CloudFront requests
                      content from the origin. For more information, see Origin Path in the Amazon CloudFront
                      Developer Guide.
                    * CustomHeaders (dict[str, Any], Optional):
                      A list of HTTP header names and values that CloudFront adds to the requests that it sends to the
                      origin.

                      For more information, see Adding Custom Headers to Origin Requests in the Amazon
                      CloudFront Developer Guide.

                      * Quantity (int): The number of custom headers, if any, for this distribution.
                      * Items (list[dict[str, Any]], Optional):
                          Optional: A list that contains one OriginCustomHeader element for each custom header that you
                          want CloudFront to forward to the origin. If Quantity is 0, omit Items.

                          * HeaderName (str): The name of a header that you want CloudFront to send to your origin.
                            For more information, see Adding Custom Headers to Origin Requests in the  Amazon
                            CloudFront Developer Guide.
                          * HeaderValue (str): The value for the header that you specified in the HeaderName field.

                    * S3OriginConfig (dict[str, Any], Optional):
                      Use this type to specify an origin that is an Amazon S3 bucket that is not configured with
                      static website hosting. To specify any other type of origin, including an Amazon S3 bucket that
                      is configured with static website hosting, use the CustomOriginConfig type instead.

                      * OriginAccessIdentity (str):
                        The CloudFront origin access identity to associate with the origin. Use an origin access
                        identity to configure the origin so that viewers can only access objects in an Amazon S3 bucket
                        through CloudFront. The format of the value is:

                        origin-access-identity/cloudfront/ID-of-origin-access-identity
                        where  ID-of-origin-access-identity  is the value that CloudFront returned in
                        the ID element when you created the origin access identity.

                        If you want viewers to be able to access objects using either the CloudFront URL or
                        the Amazon S3 URL, specify an empty OriginAccessIdentity element.

                        To delete the origin access identity from an existing distribution, update the distribution
                        configuration and include an empty OriginAccessIdentity element.

                        To replace the origin access identity, update the distribution configuration and
                        specify the new origin access identity.

                        For more information about the origin access identity,
                        see Serving Private Content through CloudFront in the Amazon CloudFront Developer Guide.

                    * CustomOriginConfig (dict[str, Any], Optional):
                      Use this type to specify an origin that is not an Amazon S3 bucket, with one exception. If the
                      Amazon S3 bucket is configured with static website hosting, use this type. If the Amazon S3
                      bucket is not configured with static website hosting, use the S3OriginConfig type instead.

                      * HTTPPort (int):
                        The HTTP port that CloudFront uses to connect to the origin. Specify the HTTP port that the
                        origin listens on.
                      * HTTPSPort (int):
                        The HTTPS port that CloudFront uses to connect to the origin. Specify the HTTPS port that the
                        origin listens on.
                      * OriginProtocolPolicy (str):
                        Specifies the protocol (HTTP or HTTPS) that CloudFront uses to connect to the origin. Valid
                        values are:

                        * http-only – CloudFront always uses HTTP to connect to the origin.
                        * match-viewer –
                          CloudFront connects to the origin using the same protocol that the viewer used to
                          connect to CloudFront.
                        * https-only – CloudFront always uses HTTPS to connect to the origin.

                      * OriginSslProtocols (dict[str, Any], Optional):
                        Specifies the minimum SSL/TLS protocol that CloudFront uses when connecting to your
                        origin over HTTPS. Valid values include SSLv3, TLSv1, TLSv1.1, and TLSv1.2.

                        For more information, see Minimum Origin SSL Protocol in the Amazon CloudFront Developer Guide.

                        * Quantity (int): The number of SSL/TLS protocols that you want to allow CloudFront to use
                          when establishing an HTTPS connection with this origin.
                        * Items (list[str]): A list that contains allowed SSL/TLS protocols for this distribution.

                      * OriginReadTimeout (int, Optional):
                        Specifies how long, in seconds, CloudFront waits for a response from the origin. This is also
                        known as the origin response timeout. The minimum timeout is 1 second, the maximum is 60
                        seconds, and the default (if you don’t specify otherwise) is 30 seconds.

                        For more information, see Origin Response Timeout in the Amazon CloudFront Developer Guide.

                      * OriginKeepaliveTimeout (int, Optional):
                        Specifies how long, in seconds, CloudFront persists its connection to the origin. The minimum
                        timeout is 1 second, the maximum is 60 seconds, and the default
                        (if you don’t specify otherwise) is 5 seconds.

                        For more information, see Origin Keep-alive Timeout in the Amazon CloudFront Developer Guide.

                    * ConnectionAttempts (int, Optional):
                      The number of times that CloudFront attempts to connect to the origin. The minimum number is 1,
                      the maximum is 3, and the default (if you don’t specify otherwise) is 3.

                      For a custom origin (including an Amazon S3 bucket that’s configured with static website hosting),
                      this value also specifies the number of times that CloudFront attempts to get a response from
                      the origin, in the case of an Origin Response Timeout.

                      For more information, see Origin Connection Attempts in the Amazon CloudFront Developer Guide.

                    * ConnectionTimeout (int, Optional):
                      The number of seconds that CloudFront waits when trying to establish a connection to the origin.
                      The minimum timeout is 1 second, the maximum is 10 seconds, and the default (if you don’t
                      specify otherwise) is 10 seconds.

                      For more information, see Origin Connection Timeout in the Amazon CloudFront Developer Guide.

                    * OriginShield (dict[str, Any], Optional):
                      CloudFront Origin Shield. Using Origin Shield can help reduce the load on your origin.

                      For more information, see Using Origin Shield in the Amazon CloudFront Developer Guide.

                      * Enabled (bool): A flag that specifies whether Origin Shield is enabled.

                        When it’s enabled, CloudFront routes all requests through Origin Shield,
                        which can help protect your origin. When it’s disabled, CloudFront might send requests directly
                        to your origin from multiple edge locations or regional edge caches.
                      * OriginShieldRegion (str, Optional): The Amazon Web Services Region for Origin Shield.

                        Specify the Amazon Web Services Region that has the lowest latency to your origin.
                        To specify a region, use the region code, not the region name.
                        For example, specify the US East (Ohio) region as us-east-2.

                        When you enable CloudFront
                        Origin Shield, you must specify the Amazon Web Services Region for Origin Shield. For the list
                        of Amazon Web Services Regions that you can specify, and for help choosing the best Region for
                        your origin, see Choosing the Amazon Web Services Region for Origin Shield in the Amazon
                        CloudFront Developer Guide.

        origin_groups (dict, Optional):
            A complex type that contains information about origin groups for this distribution.

            * Quantity (int): The number of origin groups.
            * Items (list[Dict[str, Any]], Optional):
              The items (origin groups) in a distribution.

              * Id (str): The origin group's ID.
              * FailoverCriteria (dict[str, Any]):
                A complex type that contains information about the failover criteria for an origin group.

                * StatusCodes (dict[str, Any]):
                  The status codes that, when returned from the primary origin, will trigger CloudFront to
                  failover to the second origin.

                  * Quantity (int): The number of status codes.
                  * Items (list[int]): The items (status codes) for an origin group.

              * Members (dict[str, Any]): A complex type that contains information about the origins in an origin group.

                * Quantity (int): The number of origins in an origin group.
                * Items (list[dict[str, Any]]): Items (origins) in an origin group.

                  * OriginId (str): The ID for an origin in an origin group.

        default_cache_behaviour(dict):
            A complex type that describes the default cache behavior if you don't specify a CacheBehavior
            element or if files don't match any of the values of PathPattern in CacheBehavior elements. You
            must create exactly one default cache behavior.

            * TargetOriginId (str):
              The value of ID for the origin that you want CloudFront to route requests to when they use the
              default cache behavior.
            * TrustedSigners (dict[str, Any], Optional):
              AWS recommend using TrustedKeyGroups instead of TrustedSigners.

              A list of Amazon Web Services account IDs whose public keys CloudFront can use to validate signed
              URLs or signed cookies.

              When a cache behavior contains trusted signers, CloudFront requires signed URLs or signed cookies for
              all requests that match the cache behavior. The URLs or cookies must be signed with the private
              key of a CloudFront key pair in a trusted signer’s Amazon Web Services account. The signed URL
              or cookie contains information about which public key CloudFront should use to verify the
              signature. For more information, see Serving private content in the Amazon CloudFront Developer Guide.

              * Enabled (bool):
                This field is true if any of the Amazon Web Services accounts have public keys that CloudFront
                can use to verify the signatures of signed URLs and signed cookies. If not, this field is false.
              * Quantity (int): The number of Amazon Web Services accounts in the list.
              * Items (list[str], Optional): A list of Amazon Web Services account identifiers.

            * TrustedKeyGroups (dict[str, Any], Optional):
              A list of key groups that CloudFront can use to validate signed URLs or signed cookies.

              When a cache behavior contains trusted key groups, CloudFront requires signed URLs or signed cookies
              for all requests that match the cache behavior. The URLs or cookies must be signed with a
              private key whose corresponding public key is in the key group. The signed URL or cookie
              contains information about which public key CloudFront should use to verify the signature. For
              more information, see Serving private content in the Amazon CloudFront Developer Guide.

              * Enabled (bool):
                This field is true if any of the key groups in the list have public keys that CloudFront can use
                to verify the signatures of signed URLs and signed cookies. If not, this field is false.
              * Quantity (int): The number of key groups in the list.
              * Items (list[str], Optional): A list of key groups identifiers.

            * ViewerProtocolPolicy (str):
              The protocol that viewers can use to access the files in the origin specified by TargetOriginId
              when a request matches the path pattern in PathPattern. You can specify the following options:

              * allow-all: Viewers can use HTTP or HTTPS.
              * redirect-to-https: If a viewer submits an HTTP request, CloudFront returns an HTTP status code of 301
                (Moved Permanently) to the viewer along with the HTTPS URL.
                The viewer then resubmits the request using the new URL.
              * https-only:
                If a viewer sends an HTTP request, CloudFront returns an HTTP status code of 403 (Forbidden).

              For more information about requiring the HTTPS protocol, see Requiring HTTPS Between Viewers and
              CloudFront in the Amazon CloudFront Developer Guide.

              .. Note::
                  The only way to guarantee that viewers
                  retrieve an object that was fetched from the origin using HTTPS is never to use any other
                  protocol to fetch the object. If you have recently changed from HTTP to HTTPS, we recommend that
                  you clear your objects’ cache because cached objects are protocol agnostic. That means that an
                  edge location will return an object from the cache regardless of whether the current request
                  protocol matches the protocol used previously. For more information, see Managing Cache
                  Expiration in the Amazon CloudFront Developer Guide.

            * AllowedMethods (dict[str, Any], Optional):
              A complex type that controls which HTTP methods CloudFront processes and forwards to your Amazon
              S3 bucket or your custom origin. There are three choices:
              * CloudFront forwards only GET and HEAD requests.
              * CloudFront forwards only GET, HEAD, and OPTIONS requests.
              * CloudFront forwards GET, HEAD, OPTIONS, PUT, PATCH, POST, and DELETE requests.

              If you pick the third choice, you may need to restrict access to your Amazon S3 bucket or to your custom
              origin so users can't perform operations that you don't want them to. For example, you might not want
              users to have permissions to delete objects from your origin.

              * Quantity (int):
                The number of HTTP methods that you want CloudFront to forward to your origin. Valid values are
                2 (for GET and HEAD requests), 3 (for GET, HEAD, and OPTIONS requests) and 7 (for GET, HEAD,
                OPTIONS, PUT, PATCH, POST, and DELETE requests).
              * Items (list[str]):
                A complex type that contains the HTTP methods that you want CloudFront to process and forward to
                your origin.
              * CachedMethods (dict[str, Any], Optional):
                A complex type that controls whether CloudFront caches the response to requests using the
                specified HTTP methods. There are two choices:
                * CloudFront caches responses to GET and HEAD requests.
                * CloudFront caches responses to GET, HEAD, and OPTIONS requests.
                If you pick the second choice for your Amazon S3 Origin, you may need to forward
                Access-Control-Request-Method, Access-Control-Request-Headers, and Origin headers for the responses to
                be cached correctly.

                * Quantity (int):
                  The number of HTTP methods for which you want CloudFront to cache responses. Valid values are 2
                  (for caching responses to GET and HEAD requests) and 3 (for caching responses to GET, HEAD, and
                  OPTIONS requests).
                * Items (list[str]):
                  A complex type that contains the HTTP methods that you want CloudFront to cache responses to.

            * SmoothStreaming (bool, Optional):
              Indicates whether you want to distribute media files in the Microsoft Smooth Streaming format
              using the origin that is associated with this cache behavior. If so, specify true; if not,
              specify false. If you specify true for SmoothStreaming, you can still distribute other content
              using this cache behavior if the content matches the value of PathPattern.

            * Compress (bool, Optional):
              Whether you want CloudFront to automatically compress certain files for this cache behavior. If
              so, specify true; if not, specify false. For more information, see Serving Compressed Files in
              the Amazon CloudFront Developer Guide.

            * LambdaFunctionAssociations (dict[str, Any], Optional):
              A complex type that contains zero or more Lambda@Edge function associations for a cache behavior.

              * Quantity (int): The number of Lambda@Edge function associations for this cache behavior.
              * Items (list[dict[str, Any]], Optional):
                A complex type that contains LambdaFunctionAssociation items for this cache behavior.
                If Quantity is 0, you can omit Items.

                * LambdaFunctionARN (str):
                  The ARN of the Lambda@Edge function. You must specify the ARN of a function version; you can't
                  specify an alias or $LATEST.

                * EventType (str):
                  Specifies the event type that triggers a Lambda@Edge function invocation. You can specify the
                  following values:

                  * viewer-request: The function executes when CloudFront receives a request
                    from a viewer and before it checks to see whether the requested object is in the edge cache.
                  * origin-request: The function executes only when CloudFront sends a request to your origin. When
                    the requested object is in the edge cache, the function doesn't execute.
                  * origin-response: The function executes after CloudFront receives a response from the origin and
                    before it caches the object in the response. When the requested object is in the edge cache,
                    the function doesn't execute.
                  * viewer-response: The function executes before CloudFront returns the requested
                    object to the viewer. The function executes regardless of whether the object was already in the
                    edge cache. If the origin returns an HTTP status code other than HTTP 200 (OK), the function
                    doesn't execute.

                * IncludeBody (bool, Optional):
                  A flag that allows a Lambda@Edge function to have read access to the body content. For more
                  information, see Accessing the Request Body by Choosing the Include Body Option in the Amazon
                  CloudFront Developer Guide.

            * FunctionAssociations (dict[str, Any], Optional):
              A list of CloudFront functions that are associated with this cache behavior. CloudFront
              functions must be published to the LIVE stage to associate them with a cache behavior.

              * Quantity (int): The number of CloudFront functions in the list.
              * Items (list[dict[str, Any]], Optional):
                The CloudFront functions that are associated with a cache behavior in a CloudFront distribution.
                CloudFront functions must be published to the LIVE stage to associate them with a cache behavior.

                * FunctionARN (str): The Amazon Resource Name (ARN) of the function.
                * EventType (str): The event type of the function, either viewer-request or viewer-response.
                  You cannot use origin-facing event types (origin-request and origin-response)
                  with a CloudFront function.

            * FieldLevelEncryptionId (str, Optional):
              The value of ID for the field-level encryption configuration that you want CloudFront to use for
              encrypting specific fields of data for the default cache behavior.

            * RealtimeLogConfigArn (str, Optional):
              The Amazon Resource Name (ARN) of the real-time log configuration that is attached to this cache
              behavior. For more information, see Real-time logs in the Amazon CloudFront Developer Guide.

            * CachePolicyId (str, Optional):
              The unique identifier of the cache policy that is attached to the default cache behavior. For
              more information, see Creating cache policies or Using the managed cache policies in the Amazon
              CloudFront Developer Guide.

              A DefaultCacheBehavior must include either a CachePolicyId or ForwardedValues.
              We recommend that you use a CachePolicyId.

            * OriginRequestPolicyId (str, Optional):
              The unique identifier of the origin request policy that is attached to the default cache
              behavior. For more information, see Creating origin request policies or Using the managed origin
              request policies in the Amazon CloudFront Developer Guide.

            * ResponseHeadersPolicyId (str, Optional): The identifier for a response headers policy.

            * ForwardedValues (dict[str, Any], Optional):
              This field is deprecated. We recommend that you use a cache policy or an origin request policy
              instead of this field. For more information, see Working with policies in the Amazon CloudFront
              Developer Guide.

              If you want to include values in the cache key, use a cache policy. For more
              information, see Creating cache policies or Using the managed cache policies in the Amazon
              CloudFront Developer Guide.

              If you want to send values to the origin but not include them in the
              cache key, use an origin request policy. For more information, see Creating origin request
              policies or Using the managed origin request policies in the Amazon CloudFront Developer Guide.

              A DefaultCacheBehavior must include either a CachePolicyId or ForwardedValues. We recommend that
              you use a CachePolicyId. A complex type that specifies how CloudFront handles query strings,
              cookies, and HTTP headers.

              * QueryString (bool):
                This field is deprecated. We recommend that you use a cache policy or an origin request policy
                instead of this field.

                If you want to include query strings in the cache key, use a cache
                policy. For more information, see Creating cache policies in the Amazon CloudFront Developer Guide.

                If you want to send query strings to the origin but not include them in the cache key,
                use an origin request policy. For more information, see Creating origin request policies in the
                Amazon CloudFront Developer Guide.

                Indicates whether you want CloudFront to forward query
                strings to the origin that is associated with this cache behavior and cache based on the query
                string parameters. CloudFront behavior depends on the value of QueryString and on the values
                that you specify for QueryStringCacheKeys, if any:

                If you specify true for QueryString and you
                don't specify any values for QueryStringCacheKeys, CloudFront forwards all query string
                parameters to the origin and caches based on all query string parameters. Depending on how many
                query string parameters and values you have, this can adversely affect performance because
                CloudFront must forward more requests to the origin.

                If you specify true for QueryString and you
                specify one or more values for QueryStringCacheKeys, CloudFront forwards all query string
                parameters to the origin, but it only caches based on the query string parameters that you
                specify.

                If you specify false for QueryString, CloudFront doesn't forward any query string
                parameters to the origin, and doesn't cache based on query string parameters.

                For more information, see Configuring CloudFront to Cache Based on Query String Parameters in
                the Amazon CloudFront Developer Guide.

              * Cookies (dict[str, Any]):
                This field is deprecated. We recommend that you use a cache policy or an origin request policy
                instead of this field.

                If you want to include cookies in the cache key, use a cache policy. For
                more information, see Creating cache policies in the Amazon CloudFront Developer Guide.

                If you want to send cookies to the origin but not include them in the cache key, use an origin request
                policy. For more information, see Creating origin request policies in the Amazon CloudFront
                Developer Guide.

                A complex type that specifies whether you want CloudFront to forward cookies to
                the origin and, if so, which ones. For more information about forwarding cookies to the origin,
                see How CloudFront Forwards, Caches, and Logs Cookies in the Amazon CloudFront Developer Guide.

                * Forward (str):
                  This field is deprecated. We recommend that you use a cache policy or an origin request policy
                  instead of this field.

                  If you want to include cookies in the cache key, use a cache policy. For
                  more information, see Creating cache policies in the Amazon CloudFront Developer Guide.

                  If you want to send cookies to the origin but not include them in the cache key, use origin request
                  policy. For more information, see Creating origin request policies in the Amazon CloudFront
                  Developer Guide.

                  Specifies which cookies to forward to the origin for this cache behavior: all,
                  none, or the list of cookies specified in the WhitelistedNames complex type.

                  Amazon S3 doesn't  process cookies. When the cache behavior is forwarding requests to an
                  Amazon S3 origin, specify none for the Forward element.

                * WhitelistedNames (dict[str, Any], Optional):
                  This field is deprecated. We recommend that you use a cache policy or an origin request policy
                  instead of this field.

                  If you want to include cookies in the cache key, use a cache policy. For
                  more information, see Creating cache policies in the Amazon CloudFront Developer Guide.

                  If you want to send cookies to the origin but not include them in the cache key, use an origin request
                  policy. For more information, see Creating origin request policies in the Amazon CloudFront
                  Developer Guide.

                  Required if you specify whitelist for the value of Forward. A complex type that
                  specifies how many different cookies you want CloudFront to forward to the origin for this cache
                  behavior and, if you want to forward selected cookies, the names of those cookies.

                  If you specify all or none for the value of Forward, omit WhitelistedNames. If you change the value of
                  Forward from whitelist to all or none and you don't delete the WhitelistedNames element and its
                  child elements, CloudFront deletes them automatically.

                  For the current limit on the number of cookie names that you can whitelist for each cache behavior,
                  see  CloudFront Limits in the Amazon Web Services General Reference.

                  * Quantity (int): The number of cookie names in the Items list.
                  * Items (list[str], Optional): A list of cookie names.

                * Headers (dict[str, Any], Optional):
                  This field is deprecated. We recommend that you use a cache policy or an origin request policy
                  instead of this field.

                  If you want to include headers in the cache key, use a cache policy. For
                  more information, see Creating cache policies in the Amazon CloudFront Developer Guide.

                  If you want to send headers to the origin but not include them in the cache key, use an origin request
                  policy. For more information, see Creating origin request policies in the Amazon CloudFront
                  Developer Guide.

                  A complex type that specifies the Headers, if any, that you want CloudFront to
                  forward to the origin for this cache behavior (whitelisted headers). For the headers that you
                  specify, CloudFront also caches separate versions of a specified object that is based on the
                  header values in viewer requests.

                  For more information, see  Caching Content Based on Request Headers in the Amazon CloudFront Developer Guide.

                  * Quantity (int): The number of header names in the Items list.
                  * Items (list[str], Optional): A list of HTTP header names.

                * QueryStringCacheKeys (dict[str, Any], Optional):
                  This field is deprecated. We recommend that you use a cache policy or an origin request policy
                  instead of this field.

                  If you want to include query strings in the cache key, use a cache
                  policy. For more information, see Creating cache policies in the Amazon CloudFront Developer
                  Guide.

                  If you want to send query strings to the origin but not include them in the cache key,
                  use an origin request policy. For more information, see Creating origin request policies in the
                  Amazon CloudFront Developer Guide.

                  A complex type that contains information about the query
                  string parameters that you want CloudFront to use for caching for this cache behavior.

                  * Quantity (int): The number of whitelisted query string parameters for a cache behavior.
                  * Items (list[str], Optional):
                    A list that contains the query string parameters that you want CloudFront to use as a basis for
                    caching for a cache behavior. If Quantity is 0, you can omit Items.

            * MinTTL (int, Optional):
              This field is deprecated. We recommend that you use the MinTTL field in a cache policy instead
              of this field. For more information, see Creating cache policies or Using the managed cache
              policies in the Amazon CloudFront Developer Guide.

              The minimum amount of time that you want
              objects to stay in CloudFront caches before CloudFront forwards another request to your origin
              to determine whether the object has been updated. For more information, see Managing How Long
              Content Stays in an Edge Cache (Expiration) in the Amazon CloudFront Developer Guide.

              You must specify 0 for MinTTL if you configure CloudFront to forward all headers to your origin (under
              Headers, if you specify 1 for Quantity and * for Name).

            * DefaultTTL (int, Optional):
              This field is deprecated. We recommend that you use the DefaultTTL field in a cache policy
              instead of this field. For more information, see Creating cache policies or Using the managed
              cache policies in the Amazon CloudFront Developer Guide.

              The default amount of time that you
              want objects to stay in CloudFront caches before CloudFront forwards another request to your
              origin to determine whether the object has been updated. The value that you specify applies only
              when your origin does not add HTTP headers such as Cache-Control max-age, Cache-Control
              s-maxage, and Expires to objects. For more information, see Managing How Long Content Stays in
              an Edge Cache (Expiration) in the Amazon CloudFront Developer Guide.

            * MaxTTL (int, Optional):
              This field is deprecated. We recommend that you use the MaxTTL field in a cache policy instead
              of this field. For more information, see Creating cache policies or Using the managed cache
              policies in the Amazon CloudFront Developer Guide.

              The maximum amount of time that you want
              objects to stay in CloudFront caches before CloudFront forwards another request to your origin
              to determine whether the object has been updated. The value that you specify applies only when
              your origin adds HTTP headers such as Cache-Control max-age, Cache-Control s-maxage, and Expires
              to objects. For more information, see Managing How Long Content Stays in an Edge Cache
              (Expiration) in the Amazon CloudFront Developer Guide.

        cache_behaviors(dict, Optional):
            A complex type that contains zero or more CacheBehavior elements.

            * Quantity (int): The number of cache behaviors for this distribution.
            * Items (list[dict[str, Any]], Optional):
              A complex type that contains cache behaviors for this distribution. If Quantity is 0, you can omit Items.

              * PathPattern (str):
                The pattern (for example, images/*.jpg) that specifies which requests to apply the behavior to.
                When CloudFront receives a viewer request, the requested path is compared with path patterns in
                the order in which cache behaviors are listed in the distribution.

                You can optionally include a slash (/) at the beginning of the path pattern. For example, /images/*.jpg.
                CloudFront behavior is the same with or without the leading /.

                The path pattern for the default cache behavior is * and cannot be changed. If the request for an object
                does not match the path pattern for any cache behaviors, CloudFront applies the behavior in the
                default cache behavior.

                For more information, see Path Pattern in the  Amazon CloudFront Developer Guide.

              * TargetOriginId (str):
                The value of ID for the origin that you want CloudFront to route requests to when they match
                this cache behavior.

              * TrustedSigners ('TrustedSigners', Optional):
                AWS recommend using TrustedKeyGroups instead of TrustedSigners.

                A list of Amazon Web Services account IDs whose public keys CloudFront can use to validate signed URLs
                or signed cookies.

                When a cache behavior contains trusted signers, CloudFront requires signed URLs or signed cookies for
                all requests that match the cache behavior. The URLs or cookies must be signed with the private
                key of a CloudFront key pair in the trusted signer’s Amazon Web Services account. The signed URL
                or cookie contains information about which public key CloudFront should use to verify the
                signature. For more information, see Serving private content in the Amazon CloudFront Developer
                Guide.

              * TrustedKeyGroups ('TrustedKeyGroups', Optional):
                A list of key groups that CloudFront can use to validate signed URLs or signed cookies.

                When a cache behavior contains trusted key groups, CloudFront requires signed URLs or signed cookies
                for all requests that match the cache behavior. The URLs or cookies must be signed with a
                private key whose corresponding public key is in the key group. The signed URL or cookie
                contains information about which public key CloudFront should use to verify the signature. For
                more information, see Serving private content in the Amazon CloudFront Developer Guide.

              * ViewerProtocolPolicy (str):
                The protocol that viewers can use to access the files in the origin specified by TargetOriginId
                when a request matches the path pattern in PathPattern. You can specify the following options:

                * allow-all: Viewers can use HTTP or HTTPS.

                * redirect-to-https: If a viewer submits an HTTP request, CloudFront returns an HTTP status code of
                  301 (Moved Permanently) to the viewer along with the HTTPS URL. The viewer then resubmits the request
                  using the new URL.

                * https-only:
                  If a viewer sends an HTTP request, CloudFront returns an HTTP status code of 403 (Forbidden).

                For more information about requiring the HTTPS protocol, see Requiring HTTPS Between Viewers and
                CloudFront in the Amazon CloudFront Developer Guide.  The only way to guarantee that viewers
                retrieve an object that was fetched from the origin using HTTPS is never to use any other
                protocol to fetch the object. If you have recently changed from HTTP to HTTPS, we recommend that
                you clear your objects’ cache because cached objects are protocol agnostic. That means that an
                edge location will return an object from the cache regardless of whether the current request
                protocol matches the protocol used previously. For more information, see Managing Cache
                Expiration in the Amazon CloudFront Developer Guide.

              * AllowedMethods ('AllowedMethods', Optional):
                A complex type that controls which HTTP methods CloudFront processes and forwards to your Amazon
                S3 bucket or your custom origin. There are three choices:

                * CloudFront forwards only GET and HEAD requests.
                * CloudFront forwards only GET, HEAD, and OPTIONS requests.
                * CloudFront forwards GET, HEAD, OPTIONS, PUT, PATCH, POST, and DELETE requests.

                If you pick the third choice, you may need to restrict access to your Amazon S3 bucket or to your
                custom origin so users can't perform operations that you don't want them to.
                For example, you might not want users to have permissions to delete objects from your origin.

              * SmoothStreaming (bool, Optional):
                Indicates whether you want to distribute media files in the Microsoft Smooth Streaming format
                using the origin that is associated with this cache behavior. If so, specify true; if not,
                specify false. If you specify true for SmoothStreaming, you can still distribute other content
                using this cache behavior if the content matches the value of PathPattern.

              * Compress (bool, Optional):
                Whether you want CloudFront to automatically compress certain files for this cache behavior. If
                so, specify true; if not, specify false. For more information, see Serving Compressed Files in
                the Amazon CloudFront Developer Guide.

              * LambdaFunctionAssociations ('LambdaFunctionAssociations', Optional):
                A complex type that contains zero or more Lambda@Edge function associations for a cache
                behavior.

              * FunctionAssociations ('FunctionAssociations', Optional):
                A list of CloudFront functions that are associated with this cache behavior. CloudFront
                functions must be published to the LIVE stage to associate them with a cache behavior.

              * FieldLevelEncryptionId (str, Optional):
                The value of ID for the field-level encryption configuration that you want CloudFront to use for
                encrypting specific fields of data for this cache behavior.

              * RealtimeLogConfigArn (str, Optional):
                The Amazon Resource Name (ARN) of the real-time log configuration that is attached to this cache
                behavior. For more information, see Real-time logs in the Amazon CloudFront Developer Guide.

              * CachePolicyId (str, Optional):
                The unique identifier of the cache policy that is attached to this cache behavior. For more
                information, see Creating cache policies or Using the managed cache policies in the Amazon
                CloudFront Developer Guide. A CacheBehavior must include either a CachePolicyId or
                ForwardedValues. We recommend that you use a CachePolicyId.

              * OriginRequestPolicyId (str, Optional):
                The unique identifier of the origin request policy that is attached to this cache behavior. For
                more information, see Creating origin request policies or Using the managed origin request
                policies in the Amazon CloudFront Developer Guide.

              * ResponseHeadersPolicyId (str, Optional): The identifier for a response headers policy.

              * ForwardedValues ('ForwardedValues', Optional):
                This field is deprecated. We recommend that you use a cache policy or an origin request policy
                instead of this field. For more information, see Working with policies in the Amazon CloudFront
                Developer Guide.

                If you want to include values in the cache key, use a cache policy. For more
                information, see Creating cache policies or Using the managed cache policies in the Amazon
                CloudFront Developer Guide.

                If you want to send values to the origin but not include them in the
                cache key, use an origin request policy. For more information, see Creating origin request
                policies or Using the managed origin request policies in the Amazon CloudFront Developer Guide.

                A CacheBehavior must include either a CachePolicyId or ForwardedValues. We recommend that you
                use a CachePolicyId.

                A complex type that specifies how CloudFront handles query strings, cookies, and HTTP headers.

              * MinTTL (int, Optional):
                This field is deprecated. We recommend that you use the MinTTL field in a cache policy instead
                of this field. For more information, see Creating cache policies or Using the managed cache
                policies in the Amazon CloudFront Developer Guide.

                The minimum amount of time that you want
                objects to stay in CloudFront caches before CloudFront forwards another request to your origin
                to determine whether the object has been updated. For more information, see  Managing How Long
                Content Stays in an Edge Cache (Expiration) in the  Amazon CloudFront Developer Guide.

                You must specify 0 for MinTTL if you configure CloudFront to forward all headers to your origin (under
                Headers, if you specify 1 for Quantity and * for Name).

              * DefaultTTL (int, Optional):
                This field is deprecated. We recommend that you use the DefaultTTL field in a cache policy
                instead of this field. For more information, see Creating cache policies or Using the managed
                cache policies in the Amazon CloudFront Developer Guide.

                The default amount of time that you
                want objects to stay in CloudFront caches before CloudFront forwards another request to your
                origin to determine whether the object has been updated. The value that you specify applies only
                when your origin does not add HTTP headers such as Cache-Control max-age, Cache-Control
                s-maxage, and Expires to objects. For more information, see Managing How Long Content Stays in
                an Edge Cache (Expiration) in the Amazon CloudFront Developer Guide.

              * MaxTTL (int, Optional):
                This field is deprecated. We recommend that you use the MaxTTL field in
                a cache policy instead of this field. For more information, see Creating cache policies or
                Using the managed cache policies in the Amazon CloudFront Developer Guide.

                The maximum amount of time that you want objects to stay in CloudFront caches before CloudFront
                forwards another request to your origin to determine whether the object has been updated.
                The value that you specify applies only when your origin adds HTTP headers such as
                Cache-Control max-age, Cache-Control s-maxage, and Expires to objects.
                For more information, see Managing How Long Content Stays in an Edge Cache (Expiration) in
                the Amazon CloudFront Developer Guide.

        custom_error_responses(dict, Optional):
            A complex type that controls the following:

            * Whether CloudFront replaces HTTP status codes in
              the 4xx and 5xx range with custom error messages before returning the response to the viewer.
            * How long CloudFront caches HTTP status codes in the 4xx and 5xx range.

            For more information about custom error pages, see Customizing Error Responses in the Amazon CloudFront Developer
            Guide.

            * Quantity (int):
              The number of HTTP status codes for which you want to specify a custom error page and/or a
              caching duration. If Quantity is 0, you can omit Items.

            * Items (list[dict[str, Any]], Optional):
              A complex type that contains a CustomErrorResponse element for each HTTP status code for which
              you want to specify a custom error page and/or a caching duration.

              * ErrorCode (int):
                The HTTP status code for which you want to specify a custom error page and/or a caching
                duration.

              * ResponsePagePath (str, Optional):
                The path to the custom error page that you want CloudFront to return to a viewer when your
                origin returns the HTTP status code specified by ErrorCode, for example, /4xx-
                errors/403-forbidden.html. If you want to store your objects and your custom error pages in
                different locations, your distribution must include a cache behavior for which the following is
                true:

                * The value of PathPattern matches the path to your custom error messages. For example,
                  suppose you saved custom error pages for 4xx errors in an Amazon S3 bucket in a directory named
                  /4xx-errors. Your distribution must include a cache behavior for which the path pattern routes
                  requests for your custom error pages to that location, for example, /4xx-errors/*.

                * The value of TargetOriginId specifies the value of the ID element for the origin that contains
                  your custom error pages.

                If you specify a value for ResponsePagePath, you must also specify a value for
                ResponseCode. We recommend that you store custom error pages in an Amazon S3 bucket. If you
                store custom error pages on an HTTP server and the server starts to return 5xx errors,
                CloudFront can't get the files that you want to return to viewers because the origin server is
                unavailable.

              * ResponseCode (str, Optional):
                The HTTP status code that you want CloudFront to return to the viewer along with the custom
                error page. There are a variety of reasons that you might want CloudFront to return a status
                code different from the status code that your origin returned to CloudFront, for example:

                * Some Internet devices (some firewalls and corporate proxies, for example) intercept HTTP 4xx and 5xx
                  and prevent the response from being returned to the viewer. If you substitute 200, the response
                  typically won't be intercepted.

                * If you don't care about distinguishing among different client
                  errors or server errors, you can specify 400 or 500 as the ResponseCode for all 4xx or 5xx errors.

                * You might want to return a 200 status code (OK) and static website so your customers don't know
                  that your website is down. If you specify a value for ResponseCode, you must also specify
                  a value for ResponsePagePath.

              * ErrorCachingMinTTL (int, Optional):
                The minimum amount of time, in seconds, that you want CloudFront to cache the HTTP status code specified
                in ErrorCode. When this time period has elapsed, CloudFront queries your origin to see whether
                the problem that caused the error has been resolved and the requested object is now available.

                For more information, see Customizing Error Responses in the Amazon CloudFront Developer Guide.

        comment(str):
                An optional comment to describe the distribution. The comment cannot be longer than 128 characters.

        logging(dict, Optional):
            A complex type that controls whether access logs are written for the distribution.

            For more information about logging, see Access Logs in the Amazon CloudFront Developer Guide.

            * Enabled (bool): Specifies whether you want CloudFront to save access logs to an Amazon S3 bucket.
              If you don't want to enable logging when you create a distribution or if you want to disable logging
              for an existing distribution, specify false for Enabled, and specify empty Bucket and Prefix elements.
              If you specify false for Enabled but you specify values for Bucket, prefix, and IncludeCookies,
              the values are automatically deleted.

            * IncludeCookies (bool): Specifies whether you want CloudFront to include cookies in access logs,
              specify true for IncludeCookies. If you choose to include cookies in logs, CloudFront logs all
              cookies regardless of how you configure the cache behaviors for this distribution.
              If you don't want to include cookies when you create a distribution or if you want to disable include
              cookies for an existing distribution, specify false for IncludeCookies.

            * Bucket (str): The Amazon S3 bucket to store the access logs in,
              for example, myawslogbucket.s3.amazonaws.com.

            * Prefix (str): An optional string that you want CloudFront to prefix to the access log filenames for this
              distribution, for example, myprefix/. If you want to enable logging, but you don't want to
              specify a prefix, you still must include an empty Prefix element in the Logging element.

        price_class (str, Optional):
            The price class that corresponds with the maximum price that you want to pay for CloudFront
            service. If you specify PriceClass_All, CloudFront responds to requests for your objects from
            all CloudFront edge locations.

            If you specify a price class other than PriceClass_All,
            CloudFront serves your objects from the CloudFront edge location that has the lowest latency
            among the edge locations in your price class. Viewers who are in or near regions that are
            excluded from your specified price class may encounter slower performance.

            For more information
            about price classes, see Choosing the Price Class for a CloudFront Distribution in the Amazon
            CloudFront Developer Guide. For information about CloudFront pricing, including how price
            classes (such as Price Class 100) map to CloudFront regions, see Amazon CloudFront Pricing.

        enabled(bool):
            From this field, you can enable or disable the selected distribution.

        viewer_certificate(dict, Optional):
            complex type that determines the distribution’s SSL/TLS configuration for communicating with viewers.

            * CloudFrontDefaultCertificate (bool, Optional):
              If the distribution uses the CloudFront domain name such as d111111abcdef8.cloudfront.net, set
              this field to true. If the distribution uses Aliases (alternate domain names or CNAMEs), set
              this field to false and specify values for the following fields:

              * ACMCertificateArn or IAMCertificateId (specify a value for one, not both)
              * MinimumProtocolVersion
              * SSLSupportMethod

            * IAMCertificateId (str, Optional):
              If the distribution uses Aliases (alternate domain names or CNAMEs) and the SSL/TLS certificate
              is stored in Identity and Access Management (IAM), provide the ID of the IAM certificate.

              If you specify an IAM certificate ID, you must also specify values for MinimumProtocolVersion and
              SSLSupportMethod.

            * ACMCertificateArn (str, Optional):
              If the distribution uses Aliases (alternate domain names or CNAMEs) and the SSL/TLS certificate
              is stored in Certificate Manager (ACM), provide the Amazon Resource Name (ARN) of the ACM
              certificate. CloudFront only supports ACM certificates in the US East (N. Virginia) Region (us-east-1).

              If you specify an ACM certificate ARN, you must also specify values for MinimumProtocolVersion and SSLSupportMethod.

            * SSLSupportMethod (str, Optional):
              If the distribution uses Aliases (alternate domain names or CNAMEs), specify which viewers the
              distribution accepts HTTPS connections from.

              * sni-only – The distribution accepts HTTPS connections from only viewers that support server name
                indication (SNI). This is recommended. Most browsers and clients support SNI.

              * vip – The distribution accepts HTTPS connections from all viewers including those that don’t
                support SNI. This is not recommended, and results in additional monthly charges from CloudFront.

              * static-ip - Do not specify this value unless your distribution has been enabled for this feature by
                the CloudFront team. If you have a use case that requires static IP addresses for a distribution,
                contact CloudFront through the Amazon Web Services Support Center.

              If the distribution uses the CloudFront domain name such as d111111abcdef8.cloudfront.net,
              don’t set a value for this field.

            * MinimumProtocolVersion (str, Optional):
              If the distribution uses Aliases (alternate domain names or CNAMEs), specify the security policy
              that you want CloudFront to use for HTTPS connections with viewers. The security policy
              determines two settings:

              * The minimum SSL/TLS protocol that CloudFront can use to communicate with viewers.
              * The ciphers that CloudFront can use to encrypt the content that it returns to viewers.

              For more information, see Security Policy and Supported Protocols and Ciphers Between Viewers and
              CloudFront in the Amazon CloudFront Developer Guide.

              .. Note::
                  On the CloudFront console, this setting is called Security Policy.

              When you’re using SNI only (you set SSLSupportMethod to sni-only), you must specify TLSv1 or higher.

              If the distribution uses the CloudFront domain name such as d111111abcdef8.cloudfront.net
              (you set CloudFrontDefaultCertificate to true), CloudFront automatically sets the security policy to
              TLSv1 regardless of the value that you set here.

            * Certificate (str, Optional):
              This field is deprecated. Use one of the following fields instead:

              * ACMCertificateArn
              * IAMCertificateId
              * CloudFrontDefaultCertificate

            * CertificateSource (str, Optional):
              This field is deprecated. Use one of the following fields instead:

              * ACMCertificateArn
              * IAMCertificateId
              * CloudFrontDefaultCertificate

        restrictions(dict, Optional):
            A complex type that identifies ways in which you want to restrict distribution of your content.

            * GeoRestriction (dict[str, Any]):
              A complex type that controls the countries in which your content is distributed.
              CloudFront determines the location of your users using MaxMind GeoIP databases.

              * RestrictionType (str): The method that you want to use to restrict distribution of your content
                by country:

                * none: No geo restriction is enabled, meaning access to content is not restricted by client geo location.

                * blacklist: The Location elements specify the countries in which you don't want
                  CloudFront to distribute your content.

                * whitelist: The Location elements specify the countries in which you want CloudFront to
                  distribute your content.

              * Quantity (int): When geo restriction is enabled, this is the number of countries in your whitelist or
                blacklist. Otherwise, when it is not enabled, Quantity is 0, and you can omit Items.

              * Items (list[str], Optional):
                A complex type that contains a Location element for each country in which you want CloudFront
                either to distribute your content (whitelist) or not distribute your content (blacklist).

                The Location element is a two-letter, uppercase country code for a country that you want to include
                in your blacklist or whitelist. Include one Location element for each country.

                CloudFront and MaxMind both use ISO 3166 country codes. For the current list of countries and
                the corresponding codes, see ISO 3166-1-alpha-2 code on the International Organization for
                Standardization website.
                You can also refer to the country list on the CloudFront console, which includes both
                country names and codes.

        web_acl_id (str, Optional):
            A unique identifier that specifies the WAF web ACL, if any, to associate with this distribution.
            To specify a web ACL created using the latest version of WAF, use the ACL ARN, for example
            arn:aws:wafv2:us-east-1:123456789012:global/webacl/ExampleWebACL/473e64fd-f30b-4765-81a0-62ad96dd167a.
            To specify a web ACL created using WAF Classic, use the ACL ID,
            for example 473e64fd-f30b-4765-81a0-62ad96dd167a.

            WAF is a web application firewall that lets you monitor
            the HTTP and HTTPS requests that are forwarded to CloudFront, and lets you control access to
            your content. Based on conditions that you specify, such as the IP addresses that requests
            originate from or the values of query strings, CloudFront responds to requests either with the
            requested content or with an HTTP 403 status code (Forbidden). You can also configure CloudFront
            to return a custom error page when a request is blocked. For more information about WAF, see the
            WAF Developer Guide.

        http_version(str, Optional):
            Specify the maximum HTTP version that you want viewers to use to communicate with
            CloudFront. The default value for new web distributions is http2. Viewers that don't support
            HTTP/2 automatically use an earlier HTTP version.

            For viewers and CloudFront to use HTTP/2,
            viewers must support TLS 1.2 or later, and must support Server Name Identification (SNI). In
            general, configuring CloudFront to communicate with viewers using HTTP/2 reduces latency. You
            can improve performance by optimizing for HTTP/2. For more information, do an Internet search
            for "http/2 optimization."

        is_ipv6_enabled (bool, Optional):
            If you want CloudFront to respond to IPv6 DNS requests with an IPv6 address for your
            distribution, specify true. If you specify false, CloudFront responds to IPv6 DNS requests with
            the DNS response code NOERROR and with no IP addresses. This allows viewers to submit a second
            request, for an IPv4 address for your distribution.

            In general, you should enable IPv6 if you
            have users on IPv6 networks who want to access your content. However, if you're using signed
            URLs or signed cookies to restrict access to your content, and if you're using a custom policy
            that includes the IpAddress parameter to restrict the IP addresses that can access your content,
            don't enable IPv6. If you want to restrict access to some content by IP address and not restrict
            access to other content (or restrict access but not by IP address), you can create two
            distributions. For more information, see Creating a Signed URL Using a Custom Policy in the
            Amazon CloudFront Developer Guide.

            If you're using an Route 53 Amazon Web Services Integration
            alias resource record set to route traffic to your CloudFront distribution, you need to create a
            second alias resource record set when both of the following are true:

            * You enable IPv6 for the distribution
            * You're using alternate domain names in the URLs for your objects
            For more information, see Routing Traffic to an Amazon CloudFront Web Distribution by Using Your Domain
            Name in the Route 53 Amazon Web Services Integration Developer Guide.

            If you created a CNAME resource record set, either with Route 53
            Amazon Web Services Integration or with another DNS
            service, you don't need to make any changes. A CNAME record will route traffic to your
            distribution regardless of the IP address format of the viewer request.

        tags(dict, Optional):
            Dict in the format of {tag-key: tag-value}

        timeout(dict, Optional):
            Timeout configuration for create/update/deletion of AWS IAM Policy.

            * create (dict): Timeout configuration for creating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional): Customized timeout configuration containing delay and max attempts.
            * update(dict, Optional): Timeout configuration for updating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts: (int, Optional) Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        .. code-block:: sls

            [distribution-name]:
             aws.cloudfront.distribution.present:
               - name: 'string'
               - resource_id: 'string'
               - caller_reference: 'string'
               - comment: ''
               - origins:
                   Items:
                   - ConnectionAttempts: 'integer'
                     ConnectionTimeout: 'integer'
                     CustomHeaders:
                       Quantity: 'integer'
                     DomainName: 'string'
                     Id: 'string'
                     OriginPath: 'string'
                     OriginShield:
                       Enabled: 'Boolean'
                     S3OriginConfig:
                      OriginAccessIdentity: ''
                   Quantity: 'integer'
               - default_cache_behaviour:
                   AllowedMethods:
                     CachedMethods:
                       Items: 'List'
                       Quantity: 'integer'
                     Items: 'List'
                     Quantity: 'integer'
                   CachePolicyId: 'string'
                   Compress: 'Boolean'
                   FieldLevelEncryptionId: ''
                   FunctionAssociations:
                     Quantity: 'integer'
                   LambdaFunctionAssociations:
                     Quantity: 'integer'
                   SmoothStreaming: 'Boolean'
                   TargetOriginId: 'string'
                   TrustedKeyGroups:
                     Enabled: 'Boolean'
                     Quantity: 'integer'
                   TrustedSigners:
                     Enabled: 'Boolean'
                     Quantity: 'integer'
                   ViewerProtocolPolicy: 'string'
               - enabled: 'Boolean'
               - logging:
                   Enabled: 'Boolean'
                   IncludeCookies: 'Boolean'
                   Bucket: 'string'
                   Prefix: 'string'
               - viewer_certificate:
                   CertificateSource: 'string'
                   CloudFrontDefaultCertificate: 'Boolean'
                   MinimumProtocolVersion: 'string'
                   SSLSupportMethod: 'string'
               - aliases:
                   Quantity: 'integer'
               - price_class: 'string'
               - origin_groups:
                   Quantity: 'integer'
               - cache_behaviors:
                   Quantity: 'integer'
               - custom_error_responses:
                   Quantity: 'integer'
               - restrictions:
                   GeoRestriction:
                     Quantity: 'integer'
                     RestrictionType: 'string'
               - web_acl_id: 'string'
               - http_version: 'string'
               - is_ipv6_enabled: 'Boolean'
               - tags:
                   'string': 'string'
               - timeout:
                 create:
                   delay: 'integer'
                   max_attempts: 'integer'
                 update:
                   delay: 'integer'
                   max_attempts: 'integer'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            E2QWRUHAPOMQZL:
              aws.cloudfront.distribution.present:
              - name: E2QWRUHAPOMQZL
              - resource_id: E2QWRUHAPOMQZL
              - caller_reference: 7c0ca447-479b-48b2-9260-bc7a4de63478
              - comment: ''
              - origins:
                  Items:
                  - ConnectionAttempts: 3
                    ConnectionTimeout: 10
                    CustomHeaders:
                      Quantity: 0
                    DomainName: idem-test-distribution-bucket.s3.us-east-1.amazonaws.com
                    Id: idem-test-distribution-bucket.s3.us-east-1.amazonaws.com
                    OriginPath: ''
                    OriginShield:
                      Enabled: false
                    S3OriginConfig:
                      OriginAccessIdentity: ''
                  Quantity: 1
              - default_cache_behaviour:
                  AllowedMethods:
                    CachedMethods:
                      Items:
                      - HEAD
                      - GET
                      Quantity: 2
                    Items:
                    - HEAD
                    - GET
                    Quantity: 2
                  CachePolicyId: 6583eea-f89d-4fab-e63d-7e88ee58f6
                  Compress: true
                  FieldLevelEncryptionId: ''
                  FunctionAssociations:
                    Quantity: 0
                  LambdaFunctionAssociations:
                    Quantity: 0
                  SmoothStreaming: false
                  TargetOriginId: idem-test-distribution-bucket.s3.us-east-1.amazonaws.com
                  TrustedKeyGroups:
                    Enabled: false
                    Quantity: 0
                  TrustedSigners:
                    Enabled: false
                    Quantity: 0
                  ViewerProtocolPolicy: redirect-to-https
              - enabled: false
              - logging:
                  Enabled: true
                  IncludeCookies: true
                  Bucket: idem-test-distribution-bucket
                  Prefix: distribution
              - viewer_certificate:
                  CertificateSource: cloudfront
                  CloudFrontDefaultCertificate: true
                  MinimumProtocolVersion: TLSv1
                  SSLSupportMethod: vip
              - aliases:
                  Quantity: 0
              - price_class: PriceClass_All
              - origin_groups:
                  Quantity: 0
              - cache_behaviors:
                  Quantity: 0
              - custom_error_responses:
                  Quantity: 0
              - restrictions:
                  GeoRestriction:
                    Quantity: 0
                    RestrictionType: none
              - web_acl_id: ''
              - http_version: http2
              - is_ipv6_enabled: true
              - tags:
                  Name: idem-test
              - timeout:
                  create:
                    delay: 60
                    max_attempts: 35
                  update:
                    delay: 60
                    max_attempts: 35
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    plan_state = None
    if resource_id:
        before = await hub.exec.aws.cloudfront.distribution.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])

        update_ret = await hub.tool.aws.cloudfront.distribution.update_distribution(
            ctx=ctx,
            name=name,
            before=result["old_state"],
            resource_id=resource_id,
            caller_reference=caller_reference,
            origins=origins,
            default_cache_behaviour=default_cache_behaviour,
            comment=comment,
            logging=logging,
            enabled=enabled,
            viewer_certificate=viewer_certificate,
            aliases=aliases,
            price_class=price_class,
            default_root_object=default_root_object,
            origin_groups=origin_groups,
            cache_behaviors=cache_behaviors,
            custom_error_responses=custom_error_responses,
            restrictions=restrictions,
            web_acl_id=web_acl_id,
            http_version=http_version,
            is_ipv6_enabled=is_ipv6_enabled,
            if_match=before["ret"].get("ETag"),
            timeout=timeout,
        )
        result["comment"] = update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            for modified_param in update_ret["ret"]:
                plan_state[modified_param] = update_ret["ret"][modified_param]
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.cloudfront.distribution", name=name
            )

        # update tags
        if tags is not None and tags != result["old_state"].get("tags"):
            update_tags_ret = await hub.tool.aws.cloudfront.tag.update_tags(
                ctx=ctx,
                resource_arn=result["old_state"].get("arn"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_ret["comment"]
            result["result"] = result["result"] and update_tags_ret["result"]
            resource_updated = resource_updated or bool(update_tags_ret["ret"])

            if ctx.get("test", False) and update_tags_ret["ret"]:
                plan_state["tags"] = update_tags_ret["ret"]

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "caller_reference": caller_reference,
                    "origins": origins,
                    "default_cache_behaviour": default_cache_behaviour,
                    "comment": comment,
                    "logging": logging,
                    "enabled": enabled,
                    "viewer_certificate": viewer_certificate,
                    "aliases": aliases,
                    "price_class": price_class,
                    "default_root_object": default_root_object,
                    "origin_groups": origin_groups,
                    "cache_behaviors": cache_behaviors,
                    "custom_error_responses": custom_error_responses,
                    "restrictions": restrictions,
                    "web_acl_id": web_acl_id,
                    "http_version": http_version,
                    "is_ipv6_enabled": is_ipv6_enabled,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.cloudfront.distribution", name=name
            )
            return result

        distribution_config = {
            "CallerReference": caller_reference,
            "Origins": origins,
            "DefaultCacheBehavior": default_cache_behaviour,
            "Comment": comment,
            "Logging": logging,
            "Enabled": enabled,
            "ViewerCertificate": viewer_certificate,
            "Aliases": aliases,
            "PriceClass": price_class,
            "DefaultRootObject": default_root_object,
            "OriginGroups": origin_groups,
            "CacheBehaviors": cache_behaviors,
            "CustomErrorResponses": custom_error_responses,
            "Restrictions": restrictions,
            "WebACLId": web_acl_id,
            "HttpVersion": http_version,
            "IsIPV6Enabled": is_ipv6_enabled,
        }
        distribution_config = (
            hub.tool.aws.cloudfront.distribution_utils.sanitize_distribution_config(
                distribution_config
            )
        )
        ret = await hub.exec.boto3.client.cloudfront.create_distribution_with_tags(
            ctx,
            DistributionConfigWithTags={
                "DistributionConfig": distribution_config,
                "Tags": {
                    "Items": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                },
            },
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["Distribution"]["Id"]

        waiter_ret = await hub.tool.aws.cloudfront.distribution.distribution_waiter(
            ctx, name, resource_id, timeout, "create"
        )
        if not waiter_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + waiter_ret["comment"]

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.cloudfront.distribution", name=name
        )

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.aws.cloudfront.distribution.get(
                ctx, name=name, resource_id=resource_id
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "delete",
                make_dataclass(
                    "DeleteTimeout",
                    [
                        ("delay", int, field(default=60)),
                        ("max_attempts", int, field(default=35)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=60)),
                        ("max_attempts", int, field(default=35)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""Deletes a specified cloudfront distribution.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            AWS Cloudfront distribution ID.
            Idem automatically considers this resource being absent if this field is not specified.

        timeout(dict, Optional):
            Timeout configuration for create/update/deletion of AWS IAM Policy.

            * create (dict): Timeout configuration for creating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional): Customized timeout configuration containing delay and max attempts.
            * update(dict, Optional): Timeout configuration for updating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts: (int, Optional) Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        .. code-block:: sls

            [distribution-name]:
                  aws.cloudfront.distribution.absent:
                  - name: 'string'
                  - resource_id: 'string'
                  - timeout:
                      create:
                        delay: int
                        max_attempts: int
                      update:
                        delay: int
                        max_attempts: int

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            E2QWRUHAPOMQZL:
              aws.cloudfront.distribution.absent:
              - name: E2QWRUHAPOMQZL
              - resource_id: E2QWRUHAPOMQZL
              - timeout:
                  create:
                    delay: 60
                    max_attempts: 35
                  update:
                    delay: 60
                    max_attempts: 35
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudfront.distribution", name=name
        )
        return result
    before = await hub.exec.aws.cloudfront.distribution.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudfront.distribution", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.cloudfront.distribution", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        # we have to disable the distribution first for it to be deleted.
        # send an update request to disable, once disabled send delete request.

        disable_ret = await hub.tool.aws.cloudfront.distribution.disable_distribution(
            ctx=ctx,
            name=name,
            resource_id=resource_id,
            before=result["old_state"],
            if_match=before["ret"]["ETag"],
            timeout=timeout,
        )
        if not disable_ret["result"]:
            result["comment"] = disable_ret["comment"]
            result["result"] = False
            return result

        # once disabled get the latest Etag/IfMatch
        distribution_info_ret = await hub.exec.aws.cloudfront.distribution.get(
            ctx, name=name, resource_id=resource_id
        )
        if not distribution_info_ret["result"]:
            result["comment"] += distribution_info_ret["comment"]
            result["result"] = False
            return result

        ret = await hub.exec.boto3.client.cloudfront.delete_distribution(
            ctx, Id=resource_id, IfMatch=distribution_info_ret["ret"]["ETag"]
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.cloudfront.distribution", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""List CloudFront distributions.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.cloudfront.distribution
    """
    result = {}
    ret = await hub.exec.boto3.client.cloudfront.list_distributions(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe distribution {ret['comment']}")
        return {}

    for distribution in ret["ret"]["DistributionList"]["Items"]:
        resource_id = distribution.get("Id")
        # This is required to get caller reference
        distribution_config_ret = (
            await hub.exec.boto3.client.cloudfront.get_distribution(ctx, Id=resource_id)
        )
        if not distribution_config_ret["result"] or not distribution_config_ret[
            "ret"
        ].get("Distribution"):
            hub.log.debug(
                f"Could not describe distribution {resource_id} {distribution_config_ret['comment']}"
            )
            continue
        # This is required to get tags
        before_tag = await hub.exec.boto3.client.cloudfront.list_tags_for_resource(
            ctx, Resource=distribution_config_ret["ret"]["Distribution"]["ARN"]
        )
        if before_tag["result"] and before_tag["ret"].get("Tags"):
            distribution_config_ret["ret"]["Distribution"]["Tags"] = (
                before_tag["ret"].get("Tags").get("Items", [])
            )
        else:
            hub.log.debug(
                f"Could not get tags for distribution {resource_id} {before_tag['comment']}"
            )

        resource_translated = await hub.tool.aws.cloudfront.conversion_utils.convert_raw_distribution_to_present(
            ctx,
            raw_resource=distribution_config_ret["ret"]["Distribution"],
            idem_resource_name=resource_id,
        )
        result[resource_id] = {
            "aws.cloudfront.distribution.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



import copy
from typing import Any
from typing import Dict


async def update(
    hub,
    ctx,
    before: Dict[str, Any],
    detector_id: str = None,
    finding_publishing_frequency: str = None,
    data_sources: Dict[str, Any] = None,
):
    result = dict(comment=(), result=True, ret=None)
    compare_dict_result = False
    update_payload = {}
    if (finding_publishing_frequency is not None) and before["ret"].get(
        "FindingPublishingFrequency"
    ) != finding_publishing_frequency:
        update_payload["FindingPublishingFrequency"] = finding_publishing_frequency
    data_describe = before["ret"]
    data_sources_rendered = hub.tool.aws.guardduty.detector.render_data_sources(
        data_describe
    )
    if data_sources:
        compare_dict_result = hub.tool.aws.guardduty.detector.compare_data_sources(
            data_sources, data_sources_rendered
        )
    if not compare_dict_result and data_sources is not None:
        update_payload["DataSources"] = data_sources
    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.guardduty.update_detector(
                ctx=ctx, DetectorId=detector_id, **update_payload
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        if "FindingPublishingFrequency" in update_payload:
            result["ret"]["finding_publishing_frequency"] = update_payload[
                "FindingPublishingFrequency"
            ]
            result["comment"] = result["comment"] + (
                f"Update finding_publishing_frequency: {update_payload['FindingPublishingFrequency']}",
            )
        if "DataSources" in update_payload:
            result["ret"]["data_sources"] = update_payload["DataSources"]
            result["comment"] = result["comment"] + (
                f"Update data_sources: {update_payload['DataSources']}",
            )
    return result


async def update_tags(
    hub,
    ctx,
    resource_arn,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """

    Args:
        hub:
        ctx:
        resource_arn: aws resource arn
        old_tags: dict of old tags
        new_tags: dict of new tags

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """

    tags_to_add = {}
    tags_to_remove = []
    tags_result = {}
    if old_tags:
        tags_result = copy.deepcopy(old_tags)
    if new_tags is not None:
        for key, value in new_tags.items():
            if (key in old_tags and old_tags.get(key) != new_tags.get(key)) or (
                key not in old_tags
            ):
                tags_to_add[key] = value
    if old_tags:
        for key in old_tags:
            if key not in new_tags:
                tags_to_remove.append(key)
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.guardduty.untag_resource(
                ctx, ResourceArn=resource_arn, TagKeys=tags_to_remove
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
        [tags_result.pop(key, None) for key in tags_to_remove]
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.guardduty.tag_resource(
                ctx, ResourceArn=resource_arn, Tags=tags_to_add
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    tags_to_append = {**tags_result, **tags_to_add}
    result["ret"] = {"tags": tags_to_append}
    result["comment"] = (f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",)
    return result


def render_data_sources(hub, detector_describe: Dict[str, Any]) -> Dict[str, Any]:
    """
    The syntax for data_sources parameter is different in create and update scenarios. So to make it similar we are
    rendering the data  so that it can be used in update and describe functions.

    Returns:
        Dict

    """
    data_sources_new = {}
    kubernetes_source = {}
    malware_protection_source = {}
    if "DataSources" in detector_describe:
        if "Kubernetes" in detector_describe["DataSources"]:
            kubernetes_source = detector_describe["DataSources"].pop("Kubernetes")
            if kubernetes_source:
                data_sources_new["Kubernetes"] = {
                    "AuditLogs": {
                        "Enable": kubernetes_source["AuditLogs"]["Status"] == "ENABLED"
                    }
                }
        if "MalwareProtection" in detector_describe["DataSources"]:
            malware_protection_source = detector_describe["DataSources"].pop(
                "MalwareProtection"
            )
            if malware_protection_source:
                data_sources_new["MalwareProtection"] = {
                    "ScanEc2InstanceWithFindings": {
                        "EbsVolumes": {
                            "Enable": kubernetes_source["ScanEc2InstanceWithFindings"][
                                "EbsVolumes"
                            ]["Status"]
                            == "ENABLED"
                        }
                    }
                }
        for key, val in detector_describe["DataSources"].items():
            data_sources_new[key] = {}
            data_sources_new[key] = {"Enable": val["Status"] == "ENABLED"}
        if kubernetes_source:
            # This is to preserve the input value. We don't have to have this, but it is safer.
            detector_describe["DataSources"]["Kubernetes"] = kubernetes_source
        if malware_protection_source:
            # This is to preserve the input value. We don't have to have this, but it is safer.
            detector_describe["DataSources"][
                "MalwareProtection"
            ] = malware_protection_source
    return data_sources_new


def compare_data_sources(
    hub, data_sources: Dict[str, Any], data_sources_new: Dict[str, Any]
):
    """
    This functions helps in comparing two dicts.
    It compares each key value in both the dicts and return true or false based on the comparison

    Returns:
        {True|False}

    """

    for key, value in data_sources.items():
        if key in data_sources_new:
            if isinstance(data_sources[key], dict):
                if not compare_data_sources(
                    hub, data_sources[key], data_sources_new[key]
                ):
                    return False
            elif value != data_sources_new[key]:
                return False
        else:
            return False
    return True



"""States module for managing ACM certificate validations."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

AMAZON = "Amazon"
DNS = "DNS"
EMAIL = "EMAIL"
ISSUED = "ISSUED"


async def present(
    hub,
    ctx,
    name: str,
    certificate_arn: str = None,
    resource_id: str = None,
    validation_record_fqdns: List[str] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "describe",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            )
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""Validate ACM Certificate.

    Before the Amazon certificate authority (CA) can issue a certificate for your site, AWS Certificate Manager (ACM)
    must prove that you own or control all of the domain names that you specify in your request. You can choose to prove
    your ownership with either Domain Name System (DNS) validation or with email validation at the time you request a
    certificate. In case of email validation, manual email approval of ACM certificate is required. Validation applies
    only to publicly trusted certificates issued by ACM. ACM does not validate domain ownership for imported certificates
    or for certificates signed by a private CA.

    Args:
        name(str):
            An Idem name of the resource.

        certificate_arn(str, Optional):
            The Amazon Resource Name (ARN) of certificate. Either certificate_arn or resource_id is required.

        resource_id(str, Optional):
            The Amazon Resource Name (ARN) of certificate to identify the resource. Either certificate_arn or resource_id is required.

        validation_record_fqdns(list[str], Optional):
            List of FQDNs that implement the validation. Only valid for DNS validation method ACM certificates.
            If this is set, the resource can implement additional sanity checks and has an explicit dependency
            on the resource that is implementing the validation.

        timeout(dict[str, Any], Optional):
            Timeout configuration for waiting for Aws Certificate to get issued.

            * describe (dict[str, Any])
                Timeout configuration for describing certificate

                * delay (int, Optional):
                    The amount of time in seconds to wait between attempts.

                * max_attempts (int, Optional):
                    Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

          [certificate-validation-resource-id]:
            aws.acm.certificate_validation.present:
            - certificate_arn: arn:aws:acm:eu-west-2:sample_arn
            - resource_id: arn:aws:acm:eu-west-2:sample_arn
            - validation_record_fqdns:
              - abc.dp.example.net.
              - abc2.testing.example.net.
            - timeout:
              describe:
                delay: 10
                max_attempts: 20
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.acm.describe_certificate(
        ctx, CertificateArn=resource_id if resource_id else certificate_arn
    )
    if not (before and before["result"]):
        hub.log.debug(before["comment"])
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    elif not before["ret"]:
        hub.log.debug(
            f"Error describing AWS Certificate {certificate_arn}: empty output"
        )
        result["comment"] = (
            f"Error describing AWS Certificate {certificate_arn}: empty output",
        )
        return result
    else:
        old_certificate = before["ret"].get("Certificate")

        if ctx.get("test", False):
            result["new_state"] = {"issued_at": "issued_time_known_after_present"}
            result["comment"] = result["comment"] + (
                f"Would issue AWS certificate {certificate_arn}",
            )
            return result
        if old_certificate.get("Status") == ISSUED and old_certificate.get("IssuedAt"):
            old_state = {"issued_at": old_certificate.get("IssuedAt")}
            result["old_state"] = old_state
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"] = (
                f"Certificate '{certificate_arn}' issued at {old_certificate.get('IssuedAt')}",
            )
            return result

        # Check Issuer
        if old_certificate.get("Issuer") != AMAZON:
            hub.log.debug(
                f"Certificate '{certificate_arn}' is not Amazon issued, no validation necessary"
            )
            result["comment"] = (
                f"Certificate '{certificate_arn}' is not Amazon issued, no validation necessary",
            )
            result["result"] = False
            return result

        # Check Validation method
        validation_method = (
            hub.tool.aws.acm.certificate_validation_utils.extract_validation_method(
                old_certificate
            )
        )
        if validation_method != DNS:
            hub.log.debug(
                "aws.acm.certificate_validation is only valid for DNS validation"
            )
            result["comment"] = result["comment"] + (
                "aws.acm.certificate_validation is only valid for DNS validation",
            )
            result["result"] = False
            return result

        if validation_record_fqdns:
            validity_of_domain = (
                hub.tool.aws.acm.certificate_validation_utils.validation_rec(
                    validation_rec_fqdns=validation_record_fqdns,
                    certificate=old_certificate,
                )
            )
            if not validity_of_domain["result"]:
                result["comment"] = result["comment"] + validity_of_domain["comment"]
                result["result"] = False
                return result

        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=60,
            default_max_attempts=40,
            timeout_config=timeout.get("describe") if timeout else None,
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "acm",
                "certificate_validated",
                CertificateArn=certificate_arn,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False

        after = await hub.exec.boto3.client.acm.describe_certificate(
            ctx, CertificateArn=certificate_arn
        )
        after_certificate = after["ret"].get("Certificate")

        if after_certificate.get("Status") != ISSUED:
            hub.log.debug(
                f"Certificate '{certificate_arn}' not issued. It is in {after_certificate.get('Status')} state"
            )
            result["comment"] = (
                f"Certificate '{certificate_arn}' not issued It is in {after_certificate.get('Status')} state",
            )
            result["result"] = True
            return result

        result["new_state"]["issued_at"] = after_certificate.get("IssuedAt").strftime(
            "%m/%d/%Y %H:%M:%S"
        )
        result["comment"] = (
            f"Certificate '{certificate_arn}' issued at {after_certificate.get('IssuedAt')}",
        )
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """A No-Op function for certificate_validation.

    This is a configuration resource of the certificate_manager resource.
    It's not possible to delete certificate_validations,
    You can delete a certificate_manager resource by calling certificate_manager.absent,
    while providing the certificate_manager id.

    Args:
        name:
            The name of the resource.

        resource_id(str, Optional):
            The Amazon Resource Name (ARN) of certificate to identify the resource. Either certificate_arn or resource_id is required.

    Request Syntax:
        .. code-block:: sls

            [certificate-validation-resource-id]:
              aws.acm.certificate_validation.absent:
                - name: "string"

    Returns:
        Dict[str, Any]
    """
    result = dict(
        comment=(
            "No-Op: The certificate validation can not be deleted as it is not an AWS resource",
        ),
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns detailed metadata about ACM certificates required for certificates validations.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.acm.certificate_validation
    """
    result = {}
    certificate_list_ret = await hub.exec.boto3.client.acm.list_certificates(
        ctx,
        Includes={
            "keyTypes": [
                "RSA_1024",
                "RSA_2048",
                "RSA_3072",
                "RSA_4096",
                "EC_prime256v1",
                "EC_secp384r1",
                "EC_secp521r1",
            ]
        },
    )
    if not certificate_list_ret["result"]:
        hub.log.debug(
            f"Could not list AWS Certificates {certificate_list_ret['comment']}"
        )
        return {}
    certificate_list = certificate_list_ret["ret"].get("CertificateSummaryList")

    for certificate in certificate_list:

        resource_id = certificate.get("CertificateArn")
        ret = await hub.exec.boto3.client.acm.describe_certificate(
            ctx, CertificateArn=resource_id
        )
        if not ret["result"]:
            hub.log.debug(
                f"Could not describe AWS Certificate '{resource_id}' {ret['comment']}"
            )
            continue

        resource = ret["ret"].get("Certificate")

        convert_ret = (
            hub.tool.aws.acm.conversion_utils.convert_raw_acm_validation_to_present(
                raw_resource=resource,
                idem_resource_name=resource_id,
            )
        )
        resource_translated = convert_ret.get("ret")
        result[resource_id] = {
            "aws.acm.certificate_validation.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



from typing import Any
from typing import Dict
from typing import List


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: List[Dict[str, Any]],
    new_value: List[Dict[str, Any]],
    comments: List[str],
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in its "present" state

    - network_interfaces:
        - AssociatePublicIpAddress: true
          DeleteOnTermination: true
          Description: ''
          DeviceIndex: 0
          Groups:
          - sg-XXXXXXXX
          InterfaceType: interface
          Ipv6Addresses: []
          NetworkCardIndex: 0
          PrivateIpAddresses:
          - Primary: true
            PrivateIpAddress: 888.88.88.888
          SubnetId: subnet-XXXXXXXX

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    new_interfaces = {interface["DeviceIndex"]: interface for interface in new_value}
    old_interfaces = {interface["DeviceIndex"]: interface for interface in old_value}

    # create new interfaces
    interfaces_to_create = set(new_interfaces.keys()) - set(old_interfaces.keys())
    for device_index in interfaces_to_create:
        attributes = new_interfaces[device_index]

        # Create the new interface
        ret = await hub.exec.boto3.client.ec2.create_network_interface(
            ctx,
            Description=attributes.get("Description"),
            DeviceIndex=device_index,
            Groups=attributes.get("Groups"),
            InterfaceType=attributes.get("InterfaceType"),
            Ipv6Addresses=attributes.get("Ipv6Addresses"),
            PrivateIpAddresses=attributes.get("PrivateIpAddresses"),
            # Use the subnet defined in the network interface, but fallback to the one used by the instance
            SubnetId=attributes.get("SubnetId", resource.subnet_id),
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            return False

        # Attach the new interface to the instance
        interface_id = ret.ret["NetworkInterface"]["NetworkInterfaceId"]
        ret = await hub.exec.boto3.client.ec2.attach_network_interface(
            ctx,
            DeviceIndex=device_index,
            InstanceId=resource.id,
            NetworkInterfaceId=interface_id,
            NetworkCardIndex=attributes.get("NetworkCardIndex"),
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            return False

    # Delete interfaces that were removed from the state
    interfaces_to_delete = set(old_interfaces.keys()) - set(new_interfaces.keys())
    for device_index in interfaces_to_delete:
        ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(
            ctx,
            Filters=[
                {"Name": "attachment.instance-id", "Values": [resource.id]},
                {"Name": "attachment.device-index", "Values": [device_index]},
            ],
        )
        if not ret:
            comments.append(f"Interface is already absent from index: {device_index}")
            continue

        # There will only be one result
        attachment_id = next(iter(ret.ret["NetworkInterfaces"]))["Attachment"][
            "AttachmentId"
        ]

        ret = await hub.exec.boto3.client.ec2.detach_network_interface(
            ctx, AttachmentId=attachment_id, Force=False
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret:
            return False

    # Modify existing network interfaces as needed
    interfaces_to_modify = set(old_interfaces.keys()).intersection(
        new_interfaces.keys()
    )
    for device_index in interfaces_to_modify:
        if old_value[device_index] == new_value[device_index]:
            continue

        ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(
            ctx,
            Filters=[
                {"Name": "attachment.instance-id", "Values": [resource.id]},
                {"Name": "attachment.device-index", "Values": [device_index]},
            ],
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            return False

        # There will only be one result
        interface = next(iter(ret.ret["NetworkInterfaces"]))

        # Check for changes to groups
        if old_value[device_index].get("Groups") != new_value[device_index].get(
            "Groups"
        ):
            ret = await hub.exec.boto3.client.ec2.modify_network_interface_attribute(
                ctx,
                NetworkInterfaceId=interface["NetworkInterfaceId"],
                Groups=new_value[device_index].get("Groups", []),
            )
            if ret.comment:
                comments.append(ret.comment)
            if not ret.result:
                return False

        # Check for changes to description
        if old_value[device_index].get("Description") != new_value[device_index].get(
            "Description"
        ):
            ret = await hub.exec.boto3.client.ec2.modify_network_interface_attribute(
                ctx,
                NetworkInterfaceId=interface["NetworkInterfaceId"],
                Description=new_value[device_index].get("Description", ""),
            )
            if ret.comment:
                comments.append(ret.comment)
            if not ret.result:
                return False

        # Check for changes to Delete On Termination
        if old_value[device_index].get("DeleteOnTermination") != new_value[
            device_index
        ].get("DeleteOnTermination"):
            ret = await hub.exec.boto3.client.ec2.modify_network_interface_attribute(
                ctx,
                NetworkInterfaceId=interface["NetworkInterfaceId"],
                Attachment=dict(
                    AttachmentId=interface["Attachment"]["AttachmentId"],
                    DeleteOnTermination=new_value[device_index].get(
                        "DeleteOnTermination"
                    ),
                ),
            )
            if ret.comment:
                comments.append(ret.comment)
            if not ret.result:
                return False

    return True
